# sage_setup: distribution = sagemath-categories
# sage.doctest: needs sage.schemes
r"""
Jacobians in Khuri-Makdisi model

This module implements Jacobian arithmetic by Khuri-Makdisi's algorithms
[Khu2004]_ based on divisor representation by linear spaces.

Jacobian
--------

There are three models for Jacobian arithmetic by Khuri-Makdisi's algorithms.
For each of the models, one should provide a base divisor satisfying certain
degree condition. The following lists the names of the three models and the
corresponding conditions on base divisors. Let `g` be the genus of the function
field.

- ``km_large``: large model; requires an effective divisor of degree at least `2g + 1`

- ``km_medium``: medium model; requires an effective divisor of degree at least `2g + 1`

- ``km_small``: small model; requires an effective divisor of degree at least `g + 1`

To create a Jacobian in this model, specify ``'km_[large|medium|small]'`` as ``model`` and
provide a base divisor satisfying the degree condition.

EXAMPLES:

We construct a function field (of a projective curve) over a finite field::

    sage: P2.<x,y,z> = ProjectiveSpace(GF(29), 2)
    sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
    sage: C.geometric_genus()
    1
    sage: H = C.function(y/x).divisor_of_poles()
    sage: H.degree()
    3

Now we use `H` as base divisor for the large and medium models::

    sage: J_large = C.jacobian(model='km_large', base_div=H)
    sage: J_large
    Jacobian of Projective Plane Curve over Finite Field of size 29
     defined by x^3 - y^2*z + 5*z^3 (Khuri-Makdisi large model)
    sage: J_medium = C.jacobian(model='km_medium', base_div=H)
    sage: J_medium
    Jacobian of Projective Plane Curve over Finite Field of size 29
     defined by x^3 - y^2*z + 5*z^3 (Khuri-Makdisi medium model)

and for the small model, we construct an effective divisor of degree 2::

    sage: B = sum(H.support()[:2])
    sage: B.degree()
    2
    sage: J_small = C.jacobian(model='km_small', base_div=B)
    sage: J_small
    Jacobian of Projective Plane Curve over Finite Field of size 29
     defined by x^3 - y^2*z + 5*z^3 (Khuri-Makdisi small model)

Group of rational points
------------------------

The group of rational points of a Jacobian is created from the Jacobian. A
point of the Jacobian group is represented by a divisor `D - B` where `D` is
an effective divisor of the same degree with the base divisor `B`. The
divisor `D` in turn is determined by a linear subspace of the Riemann-Roch
space associated with certain multiple of `B` (depending on the model). This
allows representing points of Jacobian as matrices once we fix a basis of the
Riemann-Roch space.

EXAMPLES::

    sage: P2.<x,y,z> = ProjectiveSpace(GF(17), 2)
    sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
    sage: F = C.function_field()
    sage: H = C.function(y/x).divisor_of_poles()
    sage: J = C.jacobian(model='km_large', base_div=H)
    sage: G = J.group()
    sage: D = C([0,1,0]).place()
    sage: P1 = C([-1,2,1]).place()
    sage: P2 = C([3,7,1]).place()
    sage: p1 = G.point(P1 - D)
    sage: p2 = G.point(P2 - D)
    sage: p1
    Point of Jacobian determined by
    [ 1  0  0  0  0  0  0 12 15]
    [ 0  1  0  0  0  0  0  0 13]
    [ 0  0  1  0  0  0  0  0  2]
    [ 0  0  0  1  0  0  0  0 16]
    [ 0  0  0  0  0  1  0  0 15]
    [ 0  0  0  0  0  0  1  0  1]
    sage: p2
    Point of Jacobian determined by
    [ 1  0  0  0  0  0  0 12  5]
    [ 0  1  0  0  0  0  0  0  2]
    [ 0  0  1  0  0  0  0  0 13]
    [ 0  0  0  1  0  0  0  0  8]
    [ 0  0  0  0  0  1  0  0 10]
    [ 0  0  0  0  0  0  1  0 14]
    sage: p1 + p2
    Point of Jacobian determined by
    [ 1  0  0  0  0 16  0  5  3]
    [ 0  1  0  0  0  6  0  8 16]
    [ 0  0  1  0  0 15  0  3 10]
    [ 0  0  0  1  0  3  0  0  0]
    [ 0  0  0  0  1 12  0 16  8]
    [ 0  0  0  0  0  0  1  3  0]

AUTHORS:

- Kwankyu Lee (2022-01-24): initial version
"""

# ****************************************************************************
#       Copyright (C) 2022 Kwankyu Lee <ekwankyu@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method

from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.richcmp import op_EQ, richcmp

from sage.categories.map import Map
from sage.categories.commutative_additive_groups import CommutativeAdditiveGroups
from sage.categories.homset import Hom

from sage.matrix.constructor import matrix

from sage.combinat.integer_vector_weighted import WeightedIntegerVectors

from .place import FunctionFieldPlace
from .divisor import FunctionFieldDivisor

from .jacobian_base import (Jacobian_base,
                            JacobianGroup_base,
                            JacobianGroup_finite_field_base,
                            JacobianPoint_base,
                            JacobianPoint_finite_field_base)


class JacobianPoint(JacobianPoint_base):
    """
    Points of a Jacobian group.

    INPUT:

    - ``parent`` -- Jacobian group

    - ``w`` -- matrix

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
        sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
        sage: b = C([0,1,0]).place()
        sage: h = C.function(y/x).divisor_of_poles()
        sage: J = C.jacobian(model='km_large', base_div=h)
        sage: G = J.group()
        sage: pl = C([3,2,1]).place()
        sage: G.point(pl - b)
        Point of Jacobian determined by
        [1 0 0 0 0 0 0 2 3]
        [0 1 0 0 0 0 0 0 3]
        [0 0 1 0 0 0 0 0 1]
        [0 0 0 1 0 0 0 0 5]
        [0 0 0 0 0 1 0 0 5]
        [0 0 0 0 0 0 1 0 4]
    """
    def __init__(self, parent, w):
        """
        Initialize.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: b = C([0,1,0]).place()
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: pl = C([3,2,1]).place()
            sage: p = G.point(pl - b)
            sage: TestSuite(p).run(skip=['_test_category','_test_pickling'])
        """
        super().__init__(parent)
        w.set_immutable()
        self._w = w

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: G.zero()
            Point of Jacobian determined by
            [1 0 0 0 0 0 0 0 0]
            [0 1 0 0 0 0 0 0 0]
            [0 0 1 0 0 0 0 0 0]
            [0 0 0 0 1 0 0 0 0]
            [0 0 0 0 0 1 0 0 0]
            [0 0 0 0 0 0 0 1 0]
        """
        return f'Point of Jacobian determined by \n{self._w}'

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: F = C.function_field()
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: zero = G.zero()
            sage: {zero: 1}
            {Point of Jacobian determined by
             [1 0 0 0 0 0 0 0 0]
             [0 1 0 0 0 0 0 0 0]
             [0 0 1 0 0 0 0 0 0]
             [0 0 0 0 1 0 0 0 0]
             [0 0 0 0 0 1 0 0 0]
             [0 0 0 0 0 0 0 1 0]: 1}
        """
        return hash(self._w)

    def _richcmp_(self, other, op):
        """
        Compare ``self`` with ``other`` with respect to operator ``op``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: b = C([0,1,0]).place()
            sage: pl1 = C([-1,2,1]).place()
            sage: pl2 = C([3,2,1]).place()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: p1 = G.point(pl1 - b)
            sage: p2 = G.point(pl2 - b)
            sage: p1 == p1
            True
            sage: p1 != p2
            True
            sage: p1 > p1
            False
            sage: p1 > p2
            True
            sage: p1 < p2
            False
        """
        if op is op_EQ:
            km = self.parent()._km
            return km.equal(self._w, other._w)
        else:
            return richcmp(self._w, other._w, op)

    def _add_(self, other):
        """
        Return the sum of ``self`` and ``other``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: b = C([0,1,0]).place()
            sage: pl1 = C([-1,2,1]).place()
            sage: pl2 = C([3,2,1]).place()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: p1 = G.point(pl1 - b)
            sage: p2 = G.point(pl2 - b)
            sage: p1 + p2
            Point of Jacobian determined by
            [1 0 0 0 0 0 3 1 1]
            [0 1 0 0 0 0 3 0 2]
            [0 0 1 0 0 0 0 3 0]
            [0 0 0 1 0 0 0 0 3]
            [0 0 0 0 1 0 4 0 3]
            [0 0 0 0 0 1 6 5 2]
            sage: p1 + p2 == p2 + p1
            True
        """
        G = self.parent()
        km = G._km
        return G.element_class(self.parent(), km.add(self._w, other._w))

    def _neg_(self):
        """
        Return the negative of this point.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: F = C.function_field()
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: b = F.get_place(1)
            sage: p = C([-1,2,1]).place()
            sage: pt = G.point(p - b)
            sage: -pt
            Point of Jacobian determined by
            [1 0 0 0 0 0 1 6 0]
            [0 1 0 0 0 0 2 6 3]
            [0 0 1 0 0 0 4 1 1]
            [0 0 0 1 0 0 1 4 1]
            [0 0 0 0 1 0 5 2 6]
            [0 0 0 0 0 1 3 1 3]
            sage: -(-pt) == pt
            True
        """
        G = self.parent()
        km = G._km
        return G.element_class(self.parent(), km.negate(self._w))

    def _rmul_(self, n):
        """
        Return the ``n``-th multiple of this point.

        INPUT:

        - ``n`` -- integer

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: b = C([0,1,0]).place()
            sage: pl = C([-1,2,1]).place()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: p = G.point(pl - b)
            sage: 10*(10*p) == 100*p
            True
        """
        return self.multiple(n)

    def multiple(self, n):
        """
        Return the ``n``-th multiple of this point.

        INPUT:

        - ``n`` -- integer

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: b = C([0,1,0]).place()
            sage: pl = C([-1,2,1]).place()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: p = G.point(pl - b)
            sage: p.multiple(100)
            Point of Jacobian determined by
            [1 0 0 0 0 2 0 1 1]
            [0 1 0 0 0 5 0 1 6]
            [0 0 1 0 0 2 0 6 3]
            [0 0 0 1 0 1 0 0 0]
            [0 0 0 0 1 5 0 1 4]
            [0 0 0 0 0 0 1 1 0]
        """
        G = self.parent()
        km = G._km
        return G.element_class(self.parent(), km.multiple(self._w, n))

    def addflip(self, other):
        """
        Return the addflip of this and ``other`` point.

        The addflip of two points is by definition the negative of the sum of
        the points. This operation is faster than addition in Jacobian
        arithmetic by Khuri-Makdisi algorithms.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: b = C([0,1,0]).place()
            sage: pl1 = C([-1,2,1]).place()
            sage: pl2 = C([3,2,1]).place()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: p1 = G.point(pl1 - b)
            sage: p2 = G.point(pl2 - b)
            sage: p1.addflip(p2)
            Point of Jacobian determined by
            [1 0 0 0 0 0 0 2 6]
            [0 1 0 0 0 0 0 0 3]
            [0 0 1 0 0 0 0 0 4]
            [0 0 0 1 0 0 0 0 3]
            [0 0 0 0 0 1 0 0 5]
            [0 0 0 0 0 0 1 0 2]
            sage: _ == -(p1 + p2)
            True
        """
        G = self.parent()
        km = G._km
        return G.element_class(self.parent(), km.addflip(self._w, other._w))

    def defining_matrix(self):
        """
        Return the matrix whose row span determines the effective divisor
        representing this point.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: b = C([0,1,0]).place()
            sage: pl = C([-1,2,1]).place()
            sage: p = G.point(pl - b)
            sage: p.defining_matrix()
            [1 0 0 0 0 0 0 2 5]
            [0 1 0 0 0 0 0 0 3]
            [0 0 1 0 0 0 0 0 2]
            [0 0 0 1 0 0 0 0 6]
            [0 0 0 0 0 1 0 0 5]
            [0 0 0 0 0 0 1 0 1]
        """
        return self._w

    def divisor(self):
        """
        Return the divisor representing this point.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: F = C.function_field()
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: b = F.get_place(1)
            sage: p = C([-1,2,1]).place()
            sage: pt = G.point(p - b)
            sage: G.point(pt.divisor()) == pt
            True

        ALGORITHM: Lemma 2.1 of [Khu2004]_.
        """
        G = self.parent()
        F = G._function_field
        data = [G._from_L(f).divisor() for f in self._w.rows()]
        supp = set()
        for d in data:
            supp.update(d.support())
        supp = list(supp)
        d = F.divisor_group().zero()
        for p in supp:
            d += min(d.valuation(p) for d in data) * p.divisor()
        return d + G._div_L - G._base_div


class JacobianPoint_finite_field(JacobianPoint, JacobianPoint_finite_field_base):
    pass


class JacobianGroupEmbedding(Map):
    """
    Embeddings between Jacobian groups.

    INPUT:

    - ``base_group`` -- Jacobian group over a base field

    - ``extension_group`` -- Jacobian group over an extension field

    EXAMPLES::

        sage: k = GF(5)
        sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
        sage: C = Curve(x^3 + z^3 - y^2*z, P2)
        sage: h = C.function(y/x).divisor_of_poles()
        sage: J = C.jacobian(model='km_large', base_div=h)
        sage: G1 = J.group()
        sage: K = k.extension(2)
        sage: G2 = J.group(K)
        sage: G2.coerce_map_from(G1)
        Jacobian group embedding map:
          From: Group of rational points of Jacobian
                over Finite Field of size 5 (Khuri-Makdisi large model)
          To:   Group of rational points of Jacobian
                over Finite Field in z2 of size 5^2 (Khuri-Makdisi large model)
    """
    def __init__(self, base_group, extension_group):
        """
        Initialize.

        TESTS::

            sage: k = GF(5)
            sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
            sage: C = Curve(x^3 + z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G1 = J.group()
            sage: K = k.extension(2)
            sage: G2 = J.group(K)
            sage: map = G2.coerce_map_from(G1)
            sage: TestSuite(map).run(skip=['_test_category', '_test_pickling'])
        """
        F_ext = extension_group._function_field
        K_ext = F_ext.constant_base_field()

        self._K_ext = K_ext

        Map.__init__(self, Hom(base_group, extension_group, CommutativeAdditiveGroups()))

    def _repr_type(self):
        """
        Return string representation of ``self``.

        TESTS::

            sage: k = GF(5)
            sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
            sage: C = Curve(x^3 + z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G1 = J.group()
            sage: K = k.extension(2)
            sage: G2 = J.group(K)
            sage: G2.coerce_map_from(G1)  # indirect doctest
            Jacobian group embedding map:
              From: Group of rational points of Jacobian
                    over Finite Field of size 5 (Khuri-Makdisi large model)
              To:   Group of rational points of Jacobian
                    over Finite Field in z2 of size 5^2 (Khuri-Makdisi large model)
        """
        return 'Jacobian group embedding'

    def _call_(self, x):
        """
        Conorm map from F to F_ext.

        TESTS::

            sage: k = GF(5)
            sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
            sage: C = Curve(x^3 + z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G1 = J.group()
            sage: K = k.extension(2)
            sage: G2 = J.group(K)
            sage: m = G2.coerce_map_from(G1)
            sage: m(G1.zero()) == G2.zero()
            True
        """
        w = (x._w).change_ring(self._K_ext)

        return self.codomain().element_class(self.codomain(), w)


class JacobianGroup(UniqueRepresentation, JacobianGroup_base):
    """
    Groups of rational points of a Jacobian.

    INPUT:

    - ``parent`` -- a Jacobian

    - ``function_field`` -- a function field

    - ``base_div`` -- an effective divisor of the function field

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
        sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
        sage: h = C.function(y/x).divisor_of_poles()
        sage: J = C.jacobian(model='km_large', base_div=h)
        sage: J.group()
        Group of rational points of Jacobian
         over Finite Field of size 7 (Khuri-Makdisi large model)
    """
    Element = JacobianPoint
    _embedding_map_class = JacobianGroupEmbedding

    def __init__(self, parent, function_field, base_div):
        """
        Initialize.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
        """
        super().__init__(parent, function_field, base_div)

        D0 = base_div

        self._V_cache = 10*[None]

        V_cache = self._V_cache

        def V(n):
            if n in V_cache:
                return V_cache[n]

            Vn, from_Vn, to_Vn = (n * D0).function_space()
            V_cache[n] = (Vn, from_Vn, to_Vn)

            return Vn, from_Vn, to_Vn

        def mu(n, m, i, j):
            Vn, from_Vn, to_Vn = V(n)
            Vm, from_Vm, to_Vm = V(m)
            Vnm, from_Vnm, to_Vnm = V(n + m)
            return to_Vnm(from_Vn(Vn.gen(i)) * from_Vm(Vm.gen(j)))

        model = parent._model

        if model == 'large':
            div_L = 3 * D0
            L, from_L, to_L = V(3)
            from sage.rings.function_field.khuri_makdisi import KhuriMakdisi_large as KM
        elif model == 'medium':
            div_L = 2 * D0
            L, from_L, to_L = V(2)
            from sage.rings.function_field.khuri_makdisi import KhuriMakdisi_medium as KM
        elif model == 'small':
            div_L = 3 * D0
            L, from_L, to_L = V(3)
            from sage.rings.function_field.khuri_makdisi import KhuriMakdisi_small as KM

        self._div_L = div_L
        self._L = L
        self._from_L = from_L
        self._to_L = to_L

        w0 = self.point(function_field.divisor_group().zero()).defining_matrix()
        km = KM(lambda n: V(n)[0], mu, w0, D0.degree(), self._genus)

        self._km = km
        self._w0 = w0

        self._base_place = None

    def _repr_(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: J.group()
            Group of rational points of Jacobian
             over Finite Field of size 7 (Khuri-Makdisi large model)
        """
        r = super()._repr_()
        return r + f' (Khuri-Makdisi {self._parent._model} model)'

    def _wd_from_divisor(self, x):
        """
        Return the matrix representing the divisor ``x``.

        INPUT:

        - ``x`` -- an effective divisor

        TESTS:

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_medium', base_div=h)
            sage: G = J.group()
            sage: P = C([-1,2,1]).place()
            sage: G._wd_from_divisor(2*P)
            [1 0 0 0 4 0]
            [0 1 0 0 4 6]
            [0 0 1 0 2 1]
            [0 0 0 1 1 6]
        """
        WD = (self._div_L - x).basis_function_space()
        wd = matrix([self._to_L(f) for f in WD])
        wd.echelonize()
        return wd

    def _element_constructor_(self, x):
        """
        Construct an element of ``self`` from ``x``.

        If ``x`` is an effective divisor, then it is assumed to have the same
        degree with the base divisor.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: G(0)
            Point of Jacobian determined by
            [1 0 0 0 0 0 0 0 0]
            [0 1 0 0 0 0 0 0 0]
            [0 0 1 0 0 0 0 0 0]
            [0 0 0 0 1 0 0 0 0]
            [0 0 0 0 0 1 0 0 0]
            [0 0 0 0 0 0 0 1 0]
            sage: b = C([0,1,0]).place()
            sage: J.set_base_place(b)
            sage: p = C([-1,2,1]).place()
            sage: G(p)
            Point of Jacobian determined by
            [1 0 0 0 0 0 0 2 5]
            [0 1 0 0 0 0 0 0 3]
            [0 0 1 0 0 0 0 0 2]
            [0 0 0 1 0 0 0 0 6]
            [0 0 0 0 0 1 0 0 5]
            [0 0 0 0 0 0 1 0 1]
        """
        if x == 0:
            return self.zero()

        if isinstance(x, FunctionFieldPlace):
            if (self._base_place is not None
                and x in self._function_field.place_set()
                and x.degree() == 1):
                x = x - self._base_place
            else:
                x = x.divisor()

        if (isinstance(x, FunctionFieldDivisor)
            and x in self._function_field.divisor_group()):
            if x.degree() == 0:
                return self.point(x)
            if x.is_effective():
                wd = self._wd_from_divisor(x)
                return self.element_class(self, wd)

        raise ValueError(f"Cannot construct a point from {x}")

    def point(self, divisor):
        """
        Return the point represented by the divisor.

        INPUT:

        - ``divisor`` -- a divisor of degree zero

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: b = C([0,1,0]).place()
            sage: p = C([-1,2,1]).place()
            sage: G.point(p - b)
            Point of Jacobian determined by
            [1 0 0 0 0 0 0 2 5]
            [0 1 0 0 0 0 0 0 3]
            [0 0 1 0 0 0 0 0 2]
            [0 0 0 1 0 0 0 0 6]
            [0 0 0 0 0 1 0 0 5]
            [0 0 0 0 0 0 1 0 1]
        """
        if divisor.degree() != 0:
            raise ValueError('divisor not of degree zero')

        c = divisor + self._base_div
        f = c.basis_function_space()[0]
        d = f.divisor() + c

        wd = self._wd_from_divisor(d)
        return self.element_class(self, wd)

    @cached_method
    def zero(self):
        """
        Return the zero element of this group.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: G.zero()
            Point of Jacobian determined by
            [1 0 0 0 0 0 0 0 0]
            [0 1 0 0 0 0 0 0 0]
            [0 0 1 0 0 0 0 0 0]
            [0 0 0 0 1 0 0 0 0]
            [0 0 0 0 0 1 0 0 0]
            [0 0 0 0 0 0 0 1 0]
        """
        return self.element_class(self, self._w0)


class JacobianGroup_finite_field(JacobianGroup, JacobianGroup_finite_field_base):
    """
    Jacobian groups of function fields over finite fields.

    INPUT:

    - ``parent`` -- a Jacobian

    - ``function_field`` -- a function field

    - ``base_div`` -- an effective divisor of the function field

    EXAMPLES::

        sage: k = GF(7)
        sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
        sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
        sage: h = C.function(y/x).divisor_of_poles()
        sage: J = C.jacobian(model='km_large', base_div=h)
        sage: G1 = J.group()
        sage: K = k.extension(2)
        sage: G2 = J.group(K)
        sage: G2.coerce_map_from(G1)
        Jacobian group embedding map:
          From: Group of rational points of Jacobian
                over Finite Field of size 7 (Khuri-Makdisi large model)
          To:   Group of rational points of Jacobian
                over Finite Field in z2 of size 7^2 (Khuri-Makdisi large model)
    """
    Element = JacobianPoint_finite_field

    def __init__(self, parent, function_field, base_div):
        """
        Initialize.

        TESTS::

            sage: k = GF(7)
            sage: P2.<x,y,z> = ProjectiveSpace(k, 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: h = C.function(y/x).divisor_of_poles()
            sage: J = C.jacobian(model='km_large', base_div=h)
            sage: G = J.group()
            sage: TestSuite(G).run(skip=['_test_elements', '_test_pickling'])
        """
        super().__init__(parent, function_field, base_div)

        F = self._function_field
        K = F.constant_base_field()

        r = self._parent._function_field.constant_base_field().degree()
        frob_K = K.frobenius_endomorphism(r)

        self._frobenius_of_constant_field = frob_K

    def __iter__(self):
        """
        Return generator of points of this group.

        TESTS::

            sage: k = GF(7)
            sage: A.<x,y> = AffineSpace(k,2)
            sage: C = Curve(y^2 + x^3 + 2*x + 1).projective_closure()
            sage: b = C([0,0,1]).place()
            sage: J = C.jacobian(model='km_large', base_div=3*b)
            sage: G = J.group()
            sage: len([pt for pt in G])  # long time
            11
        """
        d0 = self._base_div.degree()
        F = self._function_field

        zero_divisor = self._km.zero_divisor()
        deg = 1
        support = []
        degrees = []
        multiples = []
        lst = []

        places_infinite = F.places_infinite()
        generators = [iter(places_infinite)]
        while True:
            while True:
                try:
                    new_pl = next(generators[-1])
                    break
                except StopIteration:
                    if deg > d0:
                        return
                    generators.append(F._places_finite(deg))
                    deg += 1
            multiples.append((d0 + 1)*[None])
            wn = self._wd_from_divisor(new_pl.divisor())
            dn = new_pl.degree()
            wr = zero_divisor
            dr = 0
            for r in range(1, d0 // dn + 1):
                wr = self._km.add_divisor(wr, wn, dr, dn)
                multiples[-1][r] = wr
                dr += dn
                for weights in WeightedIntegerVectors(d0 - dr, degrees):
                    d = dr
                    wD = wr
                    for i in range(len(support)):
                        w = weights[i]
                        if w > 0:
                            m = w * degrees[i]
                            wD = self._km.add_divisor(wD, multiples[i][w], d, m)
                            d += m
                    pt = self.element_class(self, wD)
                    if pt not in lst:
                        lst.append(pt)
                        yield pt
            support.append(new_pl)
            degrees.append(new_pl.degree())

    def _frobenius_on(self, pt):
        """
        Return the image of ``pt`` acted by the Frobenius automorphism.

        INPUT:

        - ``pt`` -- a point of ``self``

        TESTS::

            sage: k = GF(7)
            sage: A.<x,y> = AffineSpace(k,2)
            sage: C = Curve(y^2 + x^3 + 2*x + 1).projective_closure()
            sage: b = C([0,0,1]).place()
            sage: J = C.jacobian(model='km_large', base_div=3*b)
            sage: G1 = J.group()
            sage: K = k.extension(3)
            sage: G3 = J.group(K)
            sage: pt = G3.get_points(12)[-2]  # expected to be a point rational over K
            sage: pt.frobenius().frobenius().frobenius() == pt  # indirect doctest
            True
        """
        w = pt._w.apply_morphism(self._frobenius_of_constant_field)
        return self.element_class(self, w)


class Jacobian(UniqueRepresentation, Jacobian_base):
    """
    Jacobians implemented by Khuri-Makdisi's algorithms.

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
        sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
        sage: C.jacobian(model='km')
        Jacobian of Projective Plane Curve over Finite Field of size 7
         defined by x^3 - y^2*z - 2*z^3 (Khuri-Makdisi large model)
    """
    def __init__(self, function_field, base_div, model, **kwds):
        """
        Initialize.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: J = C.jacobian(model='km_large')
            sage: TestSuite(J).run(skip=['_test_elements', '_test_pickling'])

        ::

            sage: J = C.jacobian(model='km_unknown')
            Traceback (most recent call last):
            ...
            ValueError: unknown model
        """
        super().__init__(function_field, base_div, **kwds)

        if model not in ['large', 'medium', 'small']:
            raise ValueError('unknown model')

        self._model = model

        if function_field.constant_base_field().is_finite():
            self._group_class = JacobianGroup_finite_field
        else:
            self._group_class = JacobianGroup

    def _repr_(self):
        """
        Return the string representation of ``self``.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(GF(7), 2)
            sage: C = Curve(x^3 + 5*z^3 - y^2*z, P2)
            sage: C.jacobian(model='km_large')
            Jacobian of Projective Plane Curve over Finite Field of size 7
             defined by x^3 - y^2*z - 2*z^3 (Khuri-Makdisi large model)
        """
        r = super()._repr_()
        return r + f' (Khuri-Makdisi {self._model} model)'



# sage_setup: distribution = sagemath-categories
"""
(Non-negative) Integer vectors

AUTHORS:

* Mike Hansen (2007) - original module
* Nathann Cohen, David Joyner (2009-2010) - Gale-Ryser stuff
* Nathann Cohen, David Joyner (2011) - Gale-Ryser bugfix
* Travis Scrimshaw (2012-05-12) - Updated doc-strings to tell the user of
  that the class's name is a misnomer (that they only contains non-negative
  entries).
* Federico Poloni (2013) - specialized ``rank()``
* Travis Scrimshaw (2013-02-04) - Refactored to use ``ClonableIntArray``
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>,
#       Copyright (C) 2012 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from itertools import product
from collections.abc import Sequence
import numbers

from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.list_clone import ClonableArray
from sage.misc.classcall_metaclass import ClasscallMetaclass

from sage.categories.enumerated_sets import EnumeratedSets
from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.rings.infinity import PlusInfinity
from sage.arith.misc import binomial
from sage.rings.integer_ring import ZZ
from sage.rings.semirings.non_negative_integer_semiring import NN
from sage.rings.integer import Integer


def is_gale_ryser(r, s):
    r"""
    Test whether the given sequences satisfy the condition
    of the Gale-Ryser theorem.

    Given a binary matrix `B` of dimension `n\times m`, the
    vector of row sums is defined as the vector whose
    `i^{\mbox{th}}` component is equal to the sum of the `i^{\mbox{th}}`
    row in `A`. The vector of column sums is defined similarly.

    If, given a binary matrix, these two vectors are easy to compute,
    the Gale-Ryser theorem lets us decide whether, given two
    nonnegative vectors `r,s`, there exists a binary matrix
    whose row/column sums vectors are `r` and `s`.

    This functions answers accordingly.

    INPUT:

    - ``r``, ``s`` -- lists of nonnegative integers

    ALGORITHM:

    Without loss of generality, we can assume that:

    - The two given sequences do not contain any `0` ( which would
      correspond to an empty column/row )

    - The two given sequences are ordered in decreasing order
      (reordering the sequence of row (resp. column) sums amounts to
      reordering the rows (resp. columns) themselves in the matrix,
      which does not alter the columns (resp. rows) sums.

    We can then assume that `r` and `s` are partitions
    (see the corresponding class :class:`Partition`)

    If `r^*` denote the conjugate of `r`, the Gale-Ryser theorem
    asserts that a binary Matrix satisfying the constraints exists
    if and only if `s \preceq r^*`, where `\preceq` denotes
    the domination order on partitions.

    EXAMPLES::

        sage: from sage.combinat.integer_vector import is_gale_ryser
        sage: is_gale_ryser([4,2,2], [3,3,1,1])                                         # needs sage.combinat
        True
        sage: is_gale_ryser([4,2,1,1], [3,3,1,1])                                       # needs sage.combinat
        True
        sage: is_gale_ryser([3,2,1,1], [3,3,1,1])                                       # needs sage.combinat
        False

    REMARK: In the literature, what we are calling a
    Gale-Ryser sequence sometimes goes by the (rather
    generic-sounding) term ''realizable sequence''.
    """

    # The sequences only contain nonnegative integers
    if [x for x in r if x < 0] or [x for x in s if x < 0]:
        return False

    # builds the corresponding partitions, i.e.
    # removes the 0 and sorts the sequences
    from sage.combinat.partition import Partition
    r2 = Partition(sorted([x for x in r if x > 0], reverse=True))
    s2 = Partition(sorted([x for x in s if x > 0], reverse=True))

    # If the two sequences only contained zeroes
    if len(r2) == 0 and len(s2) == 0:
        return True

    rstar = Partition(r2).conjugate()

    #                                same number of 1s           domination
    return len(rstar) <= len(s2) and sum(r2) == sum(s2) and rstar.dominates(s)


def gale_ryser_theorem(p1, p2, algorithm='gale',
                       *, solver=None, integrality_tolerance=1e-3):
    r"""
    Return the binary matrix given by the Gale-Ryser theorem.

    The Gale Ryser theorem asserts that if `p_1,p_2` are two
    partitions of `n` of respective lengths `k_1,k_2`, then there is
    a binary `k_1\times k_2` matrix `M` such that `p_1` is the vector
    of row sums and `p_2` is the vector of column sums of `M`, if
    and only if the conjugate of `p_2` dominates `p_1`.

    INPUT:

    - ``p1``, ``p2`` -- list of integers representing the vectors
      of row/column sums

    - ``algorithm`` -- two possible string values:

      - ``'ryser'`` implements the construction due to Ryser [Ryser63]_.
      - ``'gale'`` (default) implements the construction due to Gale [Gale57]_.

    - ``solver`` -- (default: ``None``) specify a Mixed Integer Linear Programming
      (MILP) solver to be used. If set to ``None``, the default one is used. For
      more information on MILP solvers and which default solver is used, see
      the method
      :meth:`solve <sage.numerical.mip.MixedIntegerLinearProgram.solve>`
      of the class
      :class:`MixedIntegerLinearProgram <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``integrality_tolerance`` -- parameter for use with MILP solvers over an
      inexact base ring; see :meth:`MixedIntegerLinearProgram.get_values`

    OUTPUT: a binary matrix if it exists, ``None`` otherwise

    Gale's Algorithm:

    (Gale [Gale57]_): A matrix satisfying the constraints of its
    sums can be defined as the solution of the following
    Linear Program, which Sage knows how to solve.

    .. MATH::

        \forall i&\sum_{j=1}^{k_2} b_{i,j}=p_{1,j}\\
        \forall i&\sum_{j=1}^{k_1} b_{j,i}=p_{2,j}\\
        &b_{i,j}\mbox{ is a binary variable}

    Ryser's Algorithm:

    (Ryser [Ryser63]_): The construction of an `m \times n` matrix
    `A=A_{r,s}`, due to Ryser, is described as follows. The
    construction works if and only if have `s\preceq r^*`.

    * Construct the `m \times n` matrix `B` from `r` by defining
      the `i`-th row of `B` to be the vector whose first `r_i`
      entries are `1`, and the remainder are 0s, `1 \leq i \leq m`.
      This maximal matrix `B` with row sum `r` and ones left
      justified has column sum `r^{*}`.

    * Shift the last `1` in certain rows of `B` to column `n` in
      order to achieve the sum `s_n`.  Call this `B` again.

      * The `1`'s in column `n` are to appear in those
        rows in which `A` has the largest row sums, giving
        preference to the bottom-most positions in case of ties.
      * Note: When this step automatically "fixes" other columns,
        one must skip ahead to the first column index
        with a wrong sum in the step below.

    * Proceed inductively to construct columns `n-1`, ..., `2`, `1`.
      Note: when performing the induction on step `k`, we consider
      the row sums of the first `k` columns.

    * Set `A = B`. Return `A`.

    EXAMPLES:

    Computing the matrix for `p_1=p_2=2+2+1`::

        sage: # needs sage.combinat sage.modules sage.numerical.mip
        sage: from sage.combinat.integer_vector import gale_ryser_theorem
        sage: p1 = [2,2,1]
        sage: p2 = [2,2,1]
        sage: print(gale_ryser_theorem(p1, p2))         # not tested
        [1 1 0]
        [1 0 1]
        [0 1 0]
        sage: A = gale_ryser_theorem(p1, p2)
        sage: rs = [sum(x) for x in A.rows()]
        sage: cs = [sum(x) for x in A.columns()]
        sage: p1 == rs; p2 == cs
        True
        True

    Or for a non-square matrix with `p_1=3+3+2+1` and `p_2=3+2+2+1+1`,
    using Ryser's algorithm::

        sage: # needs sage.combinat sage.modules
        sage: from sage.combinat.integer_vector import gale_ryser_theorem
        sage: p1 = [3,3,1,1]
        sage: p2 = [3,3,1,1]
        sage: gale_ryser_theorem(p1, p2, algorithm='ryser')
        [1 1 1 0]
        [1 1 0 1]
        [1 0 0 0]
        [0 1 0 0]
        sage: p1 = [4,2,2]
        sage: p2 = [3,3,1,1]
        sage: gale_ryser_theorem(p1, p2, algorithm='ryser')
        [1 1 1 1]
        [1 1 0 0]
        [1 1 0 0]
        sage: p1 = [4,2,2,0]
        sage: p2 = [3,3,1,1,0,0]
        sage: gale_ryser_theorem(p1, p2, algorithm='ryser')
        [1 1 1 1 0 0]
        [1 1 0 0 0 0]
        [1 1 0 0 0 0]
        [0 0 0 0 0 0]
        sage: p1 = [3,3,2,1]
        sage: p2 = [3,2,2,1,1]
        sage: print(gale_ryser_theorem(p1, p2, algorithm='gale'))       # not tested
        [1 1 1 0 0]
        [1 1 0 0 1]
        [1 0 1 0 0]
        [0 0 0 1 0]

    With `0` in the sequences, and with unordered inputs::

        sage: from sage.combinat.integer_vector import gale_ryser_theorem
        sage: gale_ryser_theorem([3,3,0,1,1,0], [3,1,3,1,0], algorithm='ryser')         # needs sage.combinat sage.modules
        [1 1 1 0 0]
        [1 0 1 1 0]
        [0 0 0 0 0]
        [1 0 0 0 0]
        [0 0 1 0 0]
        [0 0 0 0 0]
        sage: p1 = [3,1,1,1,1]; p2 = [3,2,2,0]
        sage: gale_ryser_theorem(p1, p2, algorithm='ryser')                             # needs sage.combinat sage.modules
        [1 1 1 0]
        [1 0 0 0]
        [1 0 0 0]
        [0 1 0 0]
        [0 0 1 0]

    TESTS:

    This test created a random bipartite graph on `n+m` vertices. Its
    adjacency matrix is binary, and it is used to create some
    "random-looking" sequences which correspond to an existing matrix. The
    ``gale_ryser_theorem`` is then called on these sequences, and the output
    checked for correction.::

        sage: def test_algorithm(algorithm, low=10, high=50):
        ....:    n,m = randint(low,high), randint(low,high)
        ....:    g = graphs.RandomBipartite(n, m, .3)
        ....:    s1 = sorted(g.degree([(0,i) for i in range(n)]), reverse = True)
        ....:    s2 = sorted(g.degree([(1,i) for i in range(m)]), reverse = True)
        ....:    m = gale_ryser_theorem(s1, s2, algorithm = algorithm)
        ....:    ss1 = sorted(map(lambda x : sum(x) , m.rows()), reverse = True)
        ....:    ss2 = sorted(map(lambda x : sum(x) , m.columns()), reverse = True)
        ....:    if ((ss1 != s1) or (ss2 != s2)):
        ....:        print("Algorithm %s failed with this input:" % algorithm)
        ....:        print(s1, s2)

        sage: for algorithm in ["gale", "ryser"]:       # long time                     # needs sage.combinat sage.modules
        ....:    for i in range(50):
        ....:       test_algorithm(algorithm, 3, 10)

    Null matrix::

        sage: gale_ryser_theorem([0,0,0],[0,0,0,0], algorithm='gale')                   # needs sage.combinat sage.modules
        [0 0 0 0]
        [0 0 0 0]
        [0 0 0 0]
        sage: gale_ryser_theorem([0,0,0],[0,0,0,0], algorithm='ryser')                  # needs sage.combinat sage.modules
        [0 0 0 0]
        [0 0 0 0]
        [0 0 0 0]

    REFERENCES:

    ..  [Ryser63] \H. J. Ryser, Combinatorial Mathematics,
        Carus Monographs, MAA, 1963.
    ..  [Gale57] \D. Gale, A theorem on flows in networks, Pacific J. Math.
        7(1957)1073-1082.
    """
    from sage.matrix.constructor import matrix

    if not is_gale_ryser(p1, p2):
        return False

    if algorithm == "ryser":  # ryser's algorithm
        from sage.combinat.permutation import Permutation

        # Sorts the sequences if they are not, and remembers the permutation
        # applied
        tmp = sorted(enumerate(p1), reverse=True, key=lambda x: x[1])
        r = [x[1] for x in tmp]
        r_permutation = [x-1 for x in Permutation([x[0]+1 for x in tmp]).inverse()]
        m = len(r)

        tmp = sorted(enumerate(p2), reverse=True, key=lambda x: x[1])
        s = [x[1] for x in tmp]
        s_permutation = [x-1 for x in Permutation([x[0]+1 for x in tmp]).inverse()]

        # This is the partition equivalent to the sliding algorithm
        cols = []
        for t in reversed(s):
            c = [0] * m
            i = 0
            while t:
                k = i + 1
                while k < m and r[i] == r[k]:
                    k += 1
                if t >= k - i:  # == number rows of the same length
                    for j in range(i, k):
                        r[j] -= 1
                        c[j] = 1
                    t -= k - i
                else:  # Remove the t last rows of that length
                    for j in range(k-t, k):
                        r[j] -= 1
                        c[j] = 1
                    t = 0
                i = k
            cols.append(c)

        # We added columns to the back instead of the front
        A0 = matrix(list(reversed(cols))).transpose()

        # Applying the permutations to get a matrix satisfying the
        # order given by the input
        A0 = A0.matrix_from_rows_and_columns(r_permutation, s_permutation)
        return A0

    elif algorithm == "gale":
        from sage.numerical.mip import MixedIntegerLinearProgram
        k1, k2 = len(p1), len(p2)
        p = MixedIntegerLinearProgram(solver=solver)
        b = p.new_variable(binary=True)
        for (i, c) in enumerate(p1):
            p.add_constraint(p.sum([b[i, j] for j in range(k2)]) == c)
        for (i, c) in enumerate(p2):
            p.add_constraint(p.sum([b[j, i] for j in range(k1)]) == c)
        p.set_objective(None)
        p.solve()
        b = p.get_values(b, convert=ZZ, tolerance=integrality_tolerance)
        M = [[0]*k2 for i in range(k1)]
        for i in range(k1):
            for j in range(k2):
                M[i][j] = b[i, j]
        return matrix(M)

    else:
        raise ValueError('the only two algorithms available are "gale" and "ryser"')


def _default_function(l, default, i):
    """
    EXAMPLES::

        sage: from sage.combinat.integer_vector import _default_function
        sage: import functools
        sage: f = functools.partial(_default_function, [1,2,3], 99)
        sage: f(-1)
        99
        sage: f(0)
        1
        sage: f(1)
        2
        sage: f(2)
        3
        sage: f(3)
        99
    """
    try:
        if i < 0:
            return default
        return l[i]
    except IndexError:
        return default


def list2func(l, default=None):
    """
    Given a list ``l``, return a function that takes in a value ``i`` and
    return ``l[i]``. If default is not ``None``, then the function will
    return the default value for out of range ``i``'s.

    EXAMPLES::

        sage: f = sage.combinat.integer_vector.list2func([1,2,3])
        sage: f(0)
        1
        sage: f(1)
        2
        sage: f(2)
        3
        sage: f(3)
        Traceback (most recent call last):
        ...
        IndexError: list index out of range

    ::

        sage: f = sage.combinat.integer_vector.list2func([1,2,3], 0)
        sage: f(2)
        3
        sage: f(3)
        0
    """
    if default is None:
        return lambda i: l[i]
    else:
        from functools import partial
        return partial(_default_function, l, default)


class IntegerVector(ClonableArray):
    """
    An integer vector.
    """

    def check(self):
        """
        Check to make sure this is a valid integer vector by making sure
        all entries are nonnegative.

        EXAMPLES::

            sage: IV = IntegerVectors()
            sage: elt = IV([1,2,1])
            sage: elt.check()

        Check :issue:`34510`::

            sage: IV3 = IntegerVectors(n=3)
            sage: IV3([2,2])
            Traceback (most recent call last):
            ...
            ValueError: [2, 2] doesn't satisfy correct constraints
            sage: IVk3 = IntegerVectors(k=3)
            sage: IVk3([2,2])
            Traceback (most recent call last):
            ...
            ValueError: [2, 2] doesn't satisfy correct constraints
            sage: IV33 = IntegerVectors(n=3, k=3)
            sage: IV33([2,2])
            Traceback (most recent call last):
            ...
            ValueError: [2, 2] doesn't satisfy correct constraints
        """
        if any(x < 0 for x in self):
            raise ValueError("all entries must be nonnegative")
        if self not in self.parent():
            raise ValueError(f"{self} doesn't satisfy correct constraints")

    def trim(self):
        """
        Remove trailing zeros from the integer vector.

        EXAMPLES::

            sage: IV = IntegerVectors()
            sage: IV([5,3,5,1,0,0]).trim()
            [5, 3, 5, 1]
            sage: IV([5,0,5,1,0]).trim()
            [5, 0, 5, 1]
            sage: IV([4,3,3]).trim()
            [4, 3, 3]
            sage: IV([0,0,0]).trim()
            []

            sage: IV = IntegerVectors(k=4)
            sage: v = IV([4,3,2,0]).trim(); v
            [4, 3, 2]
            sage: v.parent()
            Integer vectors
        """
        P = IntegerVectors()
        v = list(self)
        if all(i == 0 for i in v):
            return P.element_class(P, [], check=False)
        while not v[-1]:
            v = v[:-1]
        return P.element_class(P, v, check=False)

    def specht_module(self, base_ring=None):
        r"""
        Return the Specht module corresponding to ``self``.

        EXAMPLES::

            sage: SM = IntegerVectors()([2,0,1,0,2]).specht_module(QQ); SM              # needs sage.combinat sage.modules
            Specht module of [(0, 0), (0, 1), (2, 0), (4, 0), (4, 1)] over Rational Field
            sage: s = SymmetricFunctions(QQ).s()                                        # needs sage.combinat sage.modules
            sage: s(SM.frobenius_image())                                               # needs sage.combinat sage.modules
            s[2, 2, 1]
        """
        from sage.combinat.specht_module import SpechtModule
        from sage.combinat.symmetric_group_algebra import SymmetricGroupAlgebra
        if base_ring is None:
            from sage.rings.rational_field import QQ
            base_ring = QQ
        R = SymmetricGroupAlgebra(base_ring, sum(self))
        return SpechtModule(R, self)

    def specht_module_dimension(self, base_ring=None):
        r"""
        Return the dimension of the Specht module corresponding to ``self``.

        INPUT:

        - ``BR`` -- (default: `\QQ`) the base ring

        EXAMPLES::

            sage: IntegerVectors()([2,0,1,0,2]).specht_module_dimension()               # needs sage.combinat sage.modules
            5
            sage: IntegerVectors()([2,0,1,0,2]).specht_module_dimension(GF(2))          # needs sage.combinat sage.modules sage.rings.finite_rings
            5
        """
        from sage.combinat.specht_module import specht_module_rank
        return specht_module_rank(self, base_ring)


class IntegerVectors(Parent, metaclass=ClasscallMetaclass):
    """
    The class of (nonnegative) integer vectors.

    INPUT:

    - ``n`` -- if set to an integer, returns the combinatorial class
      of integer vectors whose sum is ``n``; if set to ``None``
      (default), no such constraint is defined

    - ``k`` -- the length of the vectors; set to ``None`` (default) if
      you do not want such a constraint

    .. NOTE::

        The entries are nonnegative integers.

    EXAMPLES:

    If ``n`` is not specified, it returns the class of all integer vectors::

        sage: IntegerVectors()
        Integer vectors
        sage: [] in IntegerVectors()
        True
        sage: [1,2,1] in IntegerVectors()
        True
        sage: [1, 0, 0] in IntegerVectors()
        True

    Entries are nonnegative::

        sage: [-1, 2] in IntegerVectors()
        False

    If ``n`` is specified, then it returns the class of all integer vectors
    which sum to ``n``::

        sage: IV3 = IntegerVectors(3); IV3
        Integer vectors that sum to 3

    Note that trailing zeros are ignored so that ``[3, 0]`` does not show
    up in the following list (since ``[3]`` does)::

        sage: IntegerVectors(3, max_length=2).list()
        [[3], [2, 1], [1, 2], [0, 3]]

    If ``n`` and ``k`` are both specified, then it returns the class
    of integer vectors that sum to ``n`` and are of length ``k``::

        sage: IV53 = IntegerVectors(5,3); IV53
        Integer vectors of length 3 that sum to 5
        sage: IV53.cardinality()
        21
        sage: IV53.first()
        [5, 0, 0]
        sage: IV53.last()
        [0, 0, 5]
        sage: IV53.random_element().parent() is IV53
        True

    Further examples::

        sage: IntegerVectors(-1, 0, min_part=1).list()
        []
        sage: IntegerVectors(-1, 2, min_part=1).list()
        []
        sage: IntegerVectors(0, 0, min_part=1).list()
        [[]]
        sage: IntegerVectors(3, 0, min_part=1).list()
        []
        sage: IntegerVectors(0, 1, min_part=1).list()
        []
        sage: IntegerVectors(2, 2, min_part=1).list()
        [[1, 1]]
        sage: IntegerVectors(2, 3, min_part=1).list()
        []
        sage: IntegerVectors(4, 2, min_part=1).list()
        [[3, 1], [2, 2], [1, 3]]

    ::

        sage: IntegerVectors(0, 3, outer=[0,0,0]).list()
        [[0, 0, 0]]
        sage: IntegerVectors(1, 3, outer=[0,0,0]).list()
        []
        sage: IntegerVectors(2, 3, outer=[0,2,0]).list()
        [[0, 2, 0]]
        sage: IntegerVectors(2, 3, outer=[1,2,1]).list()
        [[1, 1, 0], [1, 0, 1], [0, 2, 0], [0, 1, 1]]
        sage: IntegerVectors(2, 3, outer=[1,1,1]).list()
        [[1, 1, 0], [1, 0, 1], [0, 1, 1]]
        sage: IntegerVectors(2, 5, outer=[1,1,1,1,1]).list()
        [[1, 1, 0, 0, 0],
         [1, 0, 1, 0, 0],
         [1, 0, 0, 1, 0],
         [1, 0, 0, 0, 1],
         [0, 1, 1, 0, 0],
         [0, 1, 0, 1, 0],
         [0, 1, 0, 0, 1],
         [0, 0, 1, 1, 0],
         [0, 0, 1, 0, 1],
         [0, 0, 0, 1, 1]]

    ::

        sage: iv = [ IntegerVectors(n,k) for n in range(-2, 7) for k in range(7) ]
        sage: all(map(lambda x: x.cardinality() == len(x.list()), iv))
        True
        sage: essai = [[1,1,1], [2,5,6], [6,5,2]]
        sage: iv = [ IntegerVectors(x[0], x[1], max_part = x[2]-1) for x in essai ]
        sage: all(map(lambda x: x.cardinality() == len(x.list()), iv))
        True

    An example showing the same output by using IntegerListsLex::

        sage: IntegerVectors(4, max_length=2).list()
        [[4], [3, 1], [2, 2], [1, 3], [0, 4]]
        sage: list(IntegerListsLex(4, max_length=2))
        [[4], [3, 1], [2, 2], [1, 3], [0, 4]]

    .. SEEALSO::

        :class:`sage.combinat.integer_lists.invlex.IntegerListsLex`
    """
    @staticmethod
    def __classcall_private__(cls, n=None, k=None, **kwargs):
        """
        Choose the correct parent based upon input.

        EXAMPLES::

            sage: IV1 = IntegerVectors(3, 2)
            sage: IV2 = IntegerVectors(3, 2)
            sage: IV1 is IV2
            True

        TESTS::

            sage: IV2 = IntegerVectors(3, 2, length=2)
            Traceback (most recent call last):
            ...
            ValueError: k and length both specified

        :issue:`29524`::

            sage: IntegerVectors(3, 3/1)
            Traceback (most recent call last):
            ...
            TypeError: 'k' must be an integer or a tuple, got Rational
        """
        if 'length' in kwargs:
            if k is not None:
                raise ValueError("k and length both specified")
            k = kwargs.pop('length')
        if kwargs:
            return IntegerVectorsConstraints(n, k, **kwargs)

        if k is None:
            if n is None:
                return IntegerVectors_all()
            return IntegerVectors_n(n)
        if n is None:
            return IntegerVectors_k(k)

        if isinstance(k, numbers.Integral):
            return IntegerVectors_nk(n, k)
        elif isinstance(k, (tuple, list)):
            return IntegerVectors_nnondescents(n, tuple(k))
        else:
            raise TypeError("'k' must be an integer or a tuple, got {}".format(type(k).__name__))

    def __init__(self, category=None):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: IV = IntegerVectors()
            sage: TestSuite(IV).run()
        """
        if category is None:
            category = EnumeratedSets()
        Parent.__init__(self, category=category)

    def _element_constructor_(self, lst):
        """
        Construct an element of ``self`` from ``lst``.

        EXAMPLES::

            sage: IV = IntegerVectors()
            sage: elt = IV([3, 1, 0, 3, 2]); elt
            [3, 1, 0, 3, 2]
            sage: elt.parent()
            Integer vectors

            sage: IV9 = IntegerVectors(9)
            sage: elt9 = IV9(elt)
            sage: elt9.parent()
            Integer vectors that sum to 9
        """
        return self.element_class(self, lst)

    Element = IntegerVector

    def __contains__(self, x):
        """
        EXAMPLES::

            sage: [] in IntegerVectors()
            True
            sage: [3,2,2,1] in IntegerVectors()
            True
        """
        if isinstance(x, IntegerVector):
            return True

        if not isinstance(x, Sequence):
            return False

        for i in x:
            if i not in ZZ:
                return False
            if i < 0:
                return False
        return True

    def _unrank_helper(self, x, rtn):
        """
        Return the element at rank ``x`` by iterating through all integer vectors beginning with ``rtn``.

        INPUT:

        - ``x`` -- nonnegative integer
        - ``rtn`` -- list of nonnegative integers

        EXAMPLES::

            sage: IV = IntegerVectors(k=5)
            sage: IV._unrank_helper(10, [2,0,0,0,0])
            [1, 0, 0, 0, 1]

            sage: IV = IntegerVectors(n=7)
            sage: IV._unrank_helper(100, [7,0,0,0])
            [2, 0, 0, 5]

            sage: IV = IntegerVectors(n=12, k=7)
            sage: IV._unrank_helper(1000, [12,0,0,0,0,0,0])
            [5, 3, 1, 1, 1, 1, 0]
        """
        ptr = 0
        while True:
            current_rank = self.rank(rtn)
            if current_rank < x:
                rtn[ptr+1] = rtn[ptr]
                rtn[ptr] = 0
                ptr += 1
            elif current_rank > x:
                rtn[ptr] -= 1
                rtn[ptr-1] += 1
            else:
                return self._element_constructor_(rtn)


class IntegerVectors_all(UniqueRepresentation, IntegerVectors):
    """
    Class of all integer vectors.
    """

    def __init__(self):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: IV = IntegerVectors()
            sage: TestSuite(IV).run()
        """
        IntegerVectors.__init__(self, category=InfiniteEnumeratedSets())

    def _repr_(self):
        """
        EXAMPLES::

            sage: IntegerVectors()
            Integer vectors
        """
        return "Integer vectors"

    def __iter__(self):
        """
        Iterate over ``self``.

        EXAMPLES::

            sage: IV = IntegerVectors()
            sage: it = IV.__iter__()
            sage: [next(it) for x in range(10)]
            [[], [1], [2], [2, 0], [1, 1], [0, 2], [3], [3, 0], [2, 1], [1, 2]]
        """
        yield self.element_class(self, [])
        n = 1
        while True:
            for k in range(1, n + 1):
                for v in integer_vectors_nk_fast_iter(n, k):
                    yield self.element_class(self, v, check=False)
            n += 1


class IntegerVectors_n(UniqueRepresentation, IntegerVectors):
    """
    Integer vectors that sum to `n`.
    """

    def __init__(self, n):
        """
        TESTS::

            sage: IV = IntegerVectors(3)
            sage: TestSuite(IV).run()
        """
        self.n = n
        if self.n == 0:
            IntegerVectors.__init__(self, category=EnumeratedSets())
        else:
            IntegerVectors.__init__(self, category=InfiniteEnumeratedSets())

    def _repr_(self):
        """
        TESTS::

            sage: IV = IntegerVectors(3)
            sage: IV
            Integer vectors that sum to 3
        """
        return "Integer vectors that sum to {}".format(self.n)

    def __iter__(self):
        """
        Iterate over ``self``.

        EXAMPLES::

            sage: it = IntegerVectors(3).__iter__()
            sage: [next(it) for x in range(10)]
            [[3],
             [3, 0],
             [2, 1],
             [1, 2],
             [0, 3],
             [3, 0, 0],
             [2, 1, 0],
             [2, 0, 1],
             [1, 2, 0],
             [1, 1, 1]]
        """
        if not self.n:
            yield self.element_class(self, [])

        k = 1
        while True:
            for iv in integer_vectors_nk_fast_iter(self.n, k):
                yield self.element_class(self, iv, check=False)
            k += 1

    def __contains__(self, x):
        """
        EXAMPLES::

            sage: [0] in IntegerVectors(0)
            True
            sage: [3] in IntegerVectors(3)
            True
            sage: [3] in IntegerVectors(2)
            False
            sage: [3,2,2,1] in IntegerVectors(9)
            False
            sage: [3,2,2,1] in IntegerVectors(8)
            True
        """
        if not IntegerVectors.__contains__(self, x):
            return False
        return sum(x) == self.n

    def rank(self, x):
        """
        Return the rank of a given element.

        INPUT:

        - ``x`` -- list with ``sum(x) == n``

        EXAMPLES::

            sage: IntegerVectors(n=5).rank([5,0])
            1
            sage: IntegerVectors(n=5).rank([3,2])
            3
        """
        if sum(x) != self.n:
            raise ValueError("argument is not a member of IntegerVectors({},{})".format(self.n, None))

        n, k, s = self.n, len(x), 0
        r = binomial(k + n - 1, n + 1)
        for i in range(k - 1):
            s += x[k - 1 - i]
            r += binomial(s + i, i + 1)
        return r

    def unrank(self, x):
        """
        Return the element at given rank x.

        INPUT:

        - ``x`` -- integer

        EXAMPLES::

            sage: IntegerVectors(n=5).unrank(2)
            [4, 1]
            sage: IntegerVectors(n=10).unrank(10)
            [1, 9]
        """
        rtn = [self.n]
        while self.rank(rtn) <= x:
            rtn.append(0)
        rtn.pop()

        return IntegerVectors._unrank_helper(self, x, rtn)

    def cardinality(self):
        """
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: IntegerVectors(n=0).cardinality()
            1
            sage: IntegerVectors(n=10).cardinality()
            +Infinity
        """
        if self.n == 0:
            return Integer(1)
        return PlusInfinity()


class IntegerVectors_k(UniqueRepresentation, IntegerVectors):
    """
    Integer vectors of length `k`.
    """

    def __init__(self, k):
        """
        TESTS::

            sage: IV = IntegerVectors(k=2)
            sage: TestSuite(IV).run()
        """
        self.k = k
        if self.k == 0:
            IntegerVectors.__init__(self, category=EnumeratedSets())
        else:
            IntegerVectors.__init__(self, category=InfiniteEnumeratedSets())

    def _repr_(self):
        """
        TESTS::

            sage: IV = IntegerVectors(k=2)
            sage: IV
            Integer vectors of length 2
        """
        return "Integer vectors of length {}".format(self.k)

    def __iter__(self):
        """
        Iterate over ``self``.

        EXAMPLES::

            sage: it = IntegerVectors(k=2).__iter__()
            sage: [next(it) for x in range(10)]
            [[0, 0],
             [1, 0],
             [0, 1],
             [2, 0],
             [1, 1],
             [0, 2],
             [3, 0],
             [2, 1],
             [1, 2],
             [0, 3]]
        """
        n = 0
        while True:
            for iv in integer_vectors_nk_fast_iter(n, self.k):
                yield self.element_class(self, iv, check=False)
            n += 1

    def __contains__(self, x):
        """
        EXAMPLES::

            sage: [] in IntegerVectors(k=0)
            True
            sage: [3] in IntegerVectors(k=1)
            True
            sage: [3] in IntegerVectors(k=2)
            False
            sage: [3,2,2,1] in IntegerVectors(k=3)
            False
            sage: [3,2,2,1] in IntegerVectors(k=4)
            True
        """
        if not IntegerVectors.__contains__(self, x):
            return False
        return len(x) == self.k

    def rank(self, x):
        """
        Return the rank of a given element.

        INPUT:

        - ``x`` -- list with ``len(x) == k``

        EXAMPLES::

            sage: IntegerVectors(k=5).rank([0,0,0,0,0])
            0
            sage: IntegerVectors(k=5).rank([1,1,0,0,0])
            7
        """
        if len(x) != self.k:
            raise ValueError("argument is not a member of IntegerVectors({},{})".format(None, self.k))

        n, k, s = sum(x), self.k, 0
        r = binomial(n + k - 1, k)
        for i in range(k - 1):
            s += x[k - 1 - i]
            r += binomial(s + i, i + 1)
        return r

    def unrank(self, x):
        """
        Return the element at given rank x.

        INPUT:

        - ``x`` -- integer such that ``x < self.cardinality()``

        EXAMPLES::

            sage: IntegerVectors(k=5).unrank(10)
            [1, 0, 0, 0, 1]
            sage: IntegerVectors(k=5).unrank(15)
            [0, 0, 2, 0, 0]
            sage: IntegerVectors(k=0).unrank(0)
            []
        """
        if self.k == 0 and x != 0:
            raise IndexError(f"Index {x} is out of range for the IntegerVector.")
        rtn = [0]*self.k
        if self.k == 0 and x == 0:
            return rtn

        while self.rank(rtn) <= x:
            rtn[0] += 1
        rtn[0] -= 1

        return IntegerVectors._unrank_helper(self, x, rtn)

    def cardinality(self):
        """
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: IntegerVectors(k=0).cardinality()
            1
            sage: IntegerVectors(k=10).cardinality()
            +Infinity
        """
        if self.k == 0:
            return Integer(1)
        return PlusInfinity()


class IntegerVectors_nk(UniqueRepresentation, IntegerVectors):
    """
    Integer vectors of length `k` that sum to `n`.

    AUTHORS:

    - Martin Albrecht
    - Mike Hansen
    """

    def __init__(self, n, k):
        """
        TESTS::

            sage: IV = IntegerVectors(2, 3)
            sage: TestSuite(IV).run()
        """
        self.n = n
        self.k = k
        IntegerVectors.__init__(self, category=FiniteEnumeratedSets())

    def _list_rec(self, n, k):
        """
        Return a list of a exponent tuples of length ``size`` such
        that the degree of the associated monomial is `D`.

        INPUT:

        - ``n`` -- degree (must be 0)

        - ``k`` -- length of exponent tuples (must be 0)

        EXAMPLES::

            sage: IV = IntegerVectors(2,3)
            sage: list(IV._list_rec(2,3))
            [(2, 0, 0), (1, 1, 0), (1, 0, 1), (0, 2, 0), (0, 1, 1), (0, 0, 2)]
        """
        if k == 1:
            yield (n,)
            return

        for nbar in range(n + 1):
            n_diff = n - nbar
            for rest in self._list_rec(nbar, k - 1):
                yield (n_diff,) + rest

    def __iter__(self):
        """
        Iterate over ``self``.

        EXAMPLES::

            sage: IV = IntegerVectors(2, 3)
            sage: list(IV)
            [[2, 0, 0], [1, 1, 0], [1, 0, 1], [0, 2, 0], [0, 1, 1], [0, 0, 2]]
            sage: list(IntegerVectors(3, 0))
            []
            sage: list(IntegerVectors(3, 1))
            [[3]]
            sage: list(IntegerVectors(0, 1))
            [[0]]
            sage: list(IntegerVectors(0, 2))
            [[0, 0]]
            sage: list(IntegerVectors(2, 2))
            [[2, 0], [1, 1], [0, 2]]
            sage: IntegerVectors(0, 0).list()
            [[]]
            sage: IntegerVectors(1, 0).list()
            []
            sage: IntegerVectors(0, 1).list()
            [[0]]
            sage: IntegerVectors(2, 2).list()
            [[2, 0], [1, 1], [0, 2]]
            sage: IntegerVectors(-1,0).list()
            []
            sage: IntegerVectors(-1,2).list()
            []
        """
        if self.n < 0:
            return

        if not self.k:
            if not self.n:
                yield self.element_class(self, [], check=False)
            return
        elif self.k == 1:
            yield self.element_class(self, [self.n], check=False)
            return

        for nbar in range(self.n+1):
            n = self.n - nbar
            for rest in integer_vectors_nk_fast_iter(nbar, self.k - 1):
                yield self.element_class(self, [n] + rest, check=False)

    def _repr_(self):
        """
        TESTS::

            sage: IV = IntegerVectors(2,3)
            sage: IV
            Integer vectors of length 3 that sum to 2
        """
        return "Integer vectors of length {} that sum to {}".format(self.k,
                                                                    self.n)

    def __contains__(self, x):
        """
        TESTS::

            sage: IV = IntegerVectors(2, 3)
            sage: all(i in IV for i in IV)
            True
            sage: [0,1,2] in IV
            False
            sage: [2.0, 0, 0] in IV
            True
            sage: [0,1,0,1] in IV
            False
            sage: [0,1,1] in IV
            True
            sage: [-1,2,1] in IV
            False

            sage: [0] in IntegerVectors(0, 1)
            True
            sage: [] in IntegerVectors(0, 0)
            True
            sage: [] in IntegerVectors(0, 1)
            False
            sage: [] in IntegerVectors(1, 0)
            False
            sage: [3] in IntegerVectors(2, 1)
            False
            sage: [3] in IntegerVectors(3, 1)
            True
            sage: [3,2,2,1] in IntegerVectors(9, 5)
            False
            sage: [3,2,2,1] in IntegerVectors(8, 5)
            False
            sage: [3,2,2,1] in IntegerVectors(8, 4)
            True

        Check :issue:`34510`::

            sage: IV33 = IntegerVectors(n=3, k=3)
            sage: IV33([0])
            Traceback (most recent call last):
            ...
            ValueError: [0] doesn't satisfy correct constraints
        """
        if not IntegerVectors.__contains__(self, x):
            return False

        if len(x) != self.k:
            return False

        if sum(x) != self.n:
            return False

        if len(x) > 0 and min(x) < 0:
            return False

        return True

    def rank(self, x):
        """
        Return the rank of a given element.

        INPUT:

        - ``x`` -- list with ``sum(x) == n`` and ``len(x) == k``

        TESTS::

            sage: IV = IntegerVectors(4,5)
            sage: list(range(IV.cardinality())) == [IV.rank(x) for x in IV]
            True
        """
        if x not in self:
            raise ValueError("argument is not a member of IntegerVectors({},{})".format(self.n, self.k))

        k, s, r = self.k, 0, 0
        for i in range(k - 1):
            s += x[k - 1 - i]
            r += binomial(s + i, i + 1)
        return r

    def unrank(self, x):
        """
        Return the element at given rank x.

        INPUT:

        - ``x`` -- integer such that ``x < self.cardinality()``

        EXAMPLES::

            sage: IntegerVectors(4,5).unrank(30)
            [1, 0, 1, 0, 2]
            sage: IntegerVectors(2,3).unrank(5)
            [0, 0, 2]
        """
        if x >= self.cardinality():
            raise IndexError(f"Index {x} is out of range for the IntegerVector.")
        rtn = [0]*self.k
        rtn[0] = self.n
        return IntegerVectors._unrank_helper(self, x, rtn)

    def cardinality(self):
        """
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: IntegerVectors(3,5).cardinality()
            35
            sage: IntegerVectors(99, 3).cardinality()
            5050
            sage: IntegerVectors(10^9 - 1, 3).cardinality()
            500000000500000000
        """
        n, k = self.n, self.k
        return Integer(binomial(n + k - 1, n))


class IntegerVectors_nnondescents(UniqueRepresentation, IntegerVectors):
    r"""
    Integer vectors graded by two parameters.

    The grading parameters on the integer vector `v` are:

    - ``n`` -- the sum of the parts of `v`

    - ``c`` -- the non descents composition of `v`

    In other words: the length of `v` equals `c_1 + \cdots + c_k`, and `v`
    is decreasing in the consecutive blocs of length `c_1, \ldots, c_k`,

    INPUT:

    - ``n`` -- the positive integer `n`
    - ``comp`` -- the composition `c`

    Those are the integer vectors of sum `n` that are lexicographically
    maximal (for the natural left-to-right reading) in their orbit by the
    Young subgroup `S_{c_1} \times \cdots \times S_{c_k}`. In particular,
    they form a set of orbit representative of integer vectors with
    respect to this Young subgroup.
    """
    @staticmethod
    def __classcall_private__(cls, n, comp):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: IntegerVectors(4, [2,1]) is IntegerVectors(int(4), (2,1))
            True
        """
        return super().__classcall__(cls, n, tuple(comp))

    def __init__(self, n, comp):
        """
        EXAMPLES::

            sage: IV = IntegerVectors(4, [2])
            sage: TestSuite(IV).run()
        """
        self.n = n
        self.comp = comp
        IntegerVectors.__init__(self, category=FiniteEnumeratedSets())

    def _repr_(self):
        """
        EXAMPLES::

            sage: IntegerVectors(4, [2])
            Integer vectors of 4 with non-descents composition [2]
        """
        return "Integer vectors of {} with non-descents composition {}".format(self.n, list(self.comp))

    def __iter__(self):
        """
        TESTS::

            sage: IntegerVectors(0, []).list()
            [[]]
            sage: IntegerVectors(5, []).list()
            []
            sage: IntegerVectors(0, [1]).list()
            [[0]]
            sage: IntegerVectors(4, [1]).list()
            [[4]]
            sage: IntegerVectors(4, [2]).list()
            [[4, 0], [3, 1], [2, 2]]
            sage: IntegerVectors(4, [2,2]).list()
            [[4, 0, 0, 0],
             [3, 1, 0, 0],
             [2, 2, 0, 0],
             [3, 0, 1, 0],
             [2, 1, 1, 0],
             [2, 0, 2, 0],
             [2, 0, 1, 1],
             [1, 1, 2, 0],
             [1, 1, 1, 1],
             [1, 0, 3, 0],
             [1, 0, 2, 1],
             [0, 0, 4, 0],
             [0, 0, 3, 1],
             [0, 0, 2, 2]]
            sage: IntegerVectors(5, [1,1,1]).list()
            [[5, 0, 0],
             [4, 1, 0],
             [4, 0, 1],
             [3, 2, 0],
             [3, 1, 1],
             [3, 0, 2],
             [2, 3, 0],
             [2, 2, 1],
             [2, 1, 2],
             [2, 0, 3],
             [1, 4, 0],
             [1, 3, 1],
             [1, 2, 2],
             [1, 1, 3],
             [1, 0, 4],
             [0, 5, 0],
             [0, 4, 1],
             [0, 3, 2],
             [0, 2, 3],
             [0, 1, 4],
             [0, 0, 5]]
            sage: IntegerVectors(0, [2,3]).list()
            [[0, 0, 0, 0, 0]]
        """
        for iv in IntegerVectors(self.n, len(self.comp)):
            blocks = [IntegerVectors(iv[i], val, max_slope=0).list()
                      for i, val in enumerate(self.comp)]
            for parts in product(*blocks):
                res = []
                for part in parts:
                    res += part
                yield self.element_class(self, res, check=False)


class IntegerVectorsConstraints(IntegerVectors):
    """
    Class of integer vectors subject to various constraints.
    """

    def __init__(self, n=None, k=None, **constraints):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: TestSuite(IntegerVectors(min_slope=0)).run()
            sage: TestSuite(IntegerVectors(3, max_slope=0)).run()
            sage: TestSuite(IntegerVectors(3, max_length=4)).run()
            sage: TestSuite(IntegerVectors(k=2, max_part=4)).run()
            sage: TestSuite(IntegerVectors(k=2, min_part=2, max_part=4)).run()
            sage: TestSuite(IntegerVectors(3, 2, max_slope=0)).run()
        """
        self.n = n
        self.k = k
        if k is not None and self.k >= 0:
            constraints['length'] = self.k
        if 'outer' in constraints:
            constraints['ceiling'] = constraints['outer']
            del constraints['outer']
        if 'inner' in constraints:
            constraints['floor'] = constraints['inner']
            del constraints['inner']
        self.constraints = constraints

        if n is not None:
            if k is not None or 'max_length' in constraints:
                category = FiniteEnumeratedSets()
            else:
                category = EnumeratedSets()
        elif k is not None and 'max_part' in constraints:  # n is None
            category = FiniteEnumeratedSets()
        else:
            category = EnumeratedSets()
        IntegerVectors.__init__(self, category=category)  # placeholder category

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: IntegerVectors(min_slope=0)
            Integer vectors with constraints: min_slope=0

            sage: IntegerVectors(3, max_length=2)
            Integer vectors that sum to 3 with constraints: max_length=2

            sage: IntegerVectors(2, 3, min_slope=0)
            Integer vectors that sum to 2 with constraints: length=3, min_slope=0
        """
        if self.n is not None:
            base = "Integer vectors that sum to {} with constraints: ".format(self.n)
        else:
            base = "Integer vectors with constraints: "
        return base + ", ".join("{}={}".format(key, self.constraints[key])
                                for key in sorted(self.constraints))

    def __eq__(self, rhs):
        """
        EXAMPLES::

            sage: IntegerVectors(min_slope=0) == IntegerVectors(min_slope=0)
            True
            sage: IntegerVectors(2, min_slope=0) == IntegerVectors(2, min_slope=0)
            True
            sage: IntegerVectors(2, 3, min_slope=0) == IntegerVectors(2, 3, min_slope=0)
            True
        """
        if isinstance(rhs, IntegerVectorsConstraints):
            return self.n == rhs.n and self.k == rhs.k and self.constraints == rhs.constraints
        return False

    def __ne__(self, rhs):
        """
        EXAMPLES::

            sage: IntegerVectors(min_slope=0) != IntegerVectors(min_slope=3)
            True
        """
        return not self.__eq__(rhs)

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: hash(IntegerVectors(min_slope=0)) == hash(IntegerVectors(min_slope=0))
            True
            sage: hash(IntegerVectors(2, min_slope=0)) == hash(IntegerVectors(2, min_slope=0))
            True
            sage: hash(IntegerVectors(2, 3, min_slope=0)) == hash(IntegerVectors(2, 3, min_slope=0))
            True
            sage: hash(IntegerVectors(min_slope=0)) != hash(IntegerVectors(min_slope=3))
            True
        """
        return hash((self.n, self.k, tuple(self.constraints.items())))

    def __contains__(self, x):
        """
        TESTS::

            sage: [3,2,2,1] in IntegerVectors(8, 4, min_part=1)                         # needs sage.combinat
            True
            sage: [3,2,2,1] in IntegerVectors(8, 4, min_part=2)                         # needs sage.combinat
            False

            sage: [0,3,0,1,2] in IntegerVectors(6, max_length=3)                        # needs sage.combinat
            False
        """
        if isinstance(x, IntegerVector) and x.parent() is self:
            return True

        if not IntegerVectors.__contains__(self, x):
            return False

        if self.k is not None and len(x) != self.k:
            return False

        if self.n is not None and sum(x) != self.n:
            return False

        from sage.combinat.misc import check_integer_list_constraints
        return check_integer_list_constraints(x, singleton=True, **self.constraints)

    def cardinality(self):
        """
        Return the cardinality of ``self``.

        EXAMPLES::

            sage: IntegerVectors(3, 3, min_part=1).cardinality()
            1
            sage: IntegerVectors(5, 3, min_part=1).cardinality()
            6
            sage: IntegerVectors(13, 4, max_part=4).cardinality()
            20
            sage: IntegerVectors(k=4, max_part=3).cardinality()
            256
            sage: IntegerVectors(k=3, min_part=2, max_part=4).cardinality()
            27
            sage: IntegerVectors(13, 4, min_part=2, max_part=4).cardinality()
            16
        """
        if self.k is None:
            if self.n is None:
                return PlusInfinity()
            if ('max_length' not in self.constraints
                    and self.constraints.get('min_part', 0) <= 0):
                return PlusInfinity()
        elif ('max_part' in self.constraints
                and self.constraints['max_part'] != PlusInfinity()):
            if (self.n is None and len(self.constraints) == 2
                    and 'min_part' in self.constraints
                    and self.constraints['min_part'] >= 0):
                num = self.constraints['max_part'] - self.constraints['min_part'] + 1
                return Integer(num ** self.k)
            if len(self.constraints) == 1:
                m = self.constraints['max_part']
                if self.n is None:
                    return Integer((m + 1) ** self.k)
                if m >= self.n:
                    return Integer(binomial(self.n + self.k - 1, self.n))
                # do by inclusion / exclusion on the number
                # i of parts greater than m
                n, k = self.n, self.k
                return Integer(sum(
                    (-1)**i * binomial(n + k - 1 - i * (m + 1), k - 1)
                    * binomial(k, i) for i in range(self.n // (m + 1) + 1)))
        return ZZ.sum(ZZ.one() for x in self)

    def __iter__(self):
        """
        EXAMPLES::

            sage: IntegerVectors(-1, 0, min_part=1).list()
            []
            sage: IntegerVectors(-1, 2, min_part=1).list()
            []
            sage: IntegerVectors(0, 0, min_part=1).list()
            [[]]
            sage: IntegerVectors(3, 0, min_part=1).list()
            []
            sage: IntegerVectors(0, 1, min_part=1).list()
            []
            sage: IntegerVectors(2, 2, min_part=1).list()
            [[1, 1]]
            sage: IntegerVectors(2, 3, min_part=1).list()
            []
            sage: IntegerVectors(4, 2, min_part=1).list()
            [[3, 1], [2, 2], [1, 3]]

        ::

            sage: IntegerVectors(0, 3, outer=[0,0,0]).list()
            [[0, 0, 0]]
            sage: IntegerVectors(1, 3, outer=[0,0,0]).list()
            []
            sage: IntegerVectors(2, 3, outer=[0,2,0]).list()
            [[0, 2, 0]]
            sage: IntegerVectors(2, 3, outer=[1,2,1]).list()
            [[1, 1, 0], [1, 0, 1], [0, 2, 0], [0, 1, 1]]
            sage: IntegerVectors(2, 3, outer=[1,1,1]).list()
            [[1, 1, 0], [1, 0, 1], [0, 1, 1]]
            sage: IntegerVectors(2, 5, outer=[1,1,1,1,1]).list()
            [[1, 1, 0, 0, 0],
             [1, 0, 1, 0, 0],
             [1, 0, 0, 1, 0],
             [1, 0, 0, 0, 1],
             [0, 1, 1, 0, 0],
             [0, 1, 0, 1, 0],
             [0, 1, 0, 0, 1],
             [0, 0, 1, 1, 0],
             [0, 0, 1, 0, 1],
             [0, 0, 0, 1, 1]]

        ::

            sage: iv = [ IntegerVectors(n, k) for n in range(-2, 7) for k in range(7) ]
            sage: all(map(lambda x: x.cardinality() == len(x.list()), iv))
            True
            sage: essai = [[1,1,1], [2,5,6], [6,5,2]]
            sage: iv = [ IntegerVectors(x[0], x[1], max_part=x[2]-1) for x in essai ]
            sage: all(map(lambda x: x.cardinality() == len(x.list()), iv))
            True
        """
        from sage.combinat.integer_lists import IntegerListsLex

        if self.n is None:
            if self.k is not None and 'max_part' in self.constraints:
                n_list = range((self.constraints['max_part'] + 1) * self.k)
            else:
                n_list = NN
        else:
            n_list = [self.n]
        for n in n_list:
            for x in IntegerListsLex(n, check=False, **self.constraints):
                yield self.element_class(self, x, check=False)


def integer_vectors_nk_fast_iter(n, k):
    """
    A fast iterator for integer vectors of ``n`` of length ``k`` which
    yields Python lists filled with Sage Integers.

    EXAMPLES::

        sage: from sage.combinat.integer_vector import integer_vectors_nk_fast_iter
        sage: list(integer_vectors_nk_fast_iter(3, 2))
        [[3, 0], [2, 1], [1, 2], [0, 3]]
        sage: list(integer_vectors_nk_fast_iter(2, 2))
        [[2, 0], [1, 1], [0, 2]]
        sage: list(integer_vectors_nk_fast_iter(1, 2))
        [[1, 0], [0, 1]]

    We check some corner cases::

        sage: list(integer_vectors_nk_fast_iter(5, 1))
        [[5]]
        sage: list(integer_vectors_nk_fast_iter(1, 1))
        [[1]]
        sage: list(integer_vectors_nk_fast_iter(2, 0))
        []
        sage: list(integer_vectors_nk_fast_iter(0, 2))
        [[0, 0]]
        sage: list(integer_vectors_nk_fast_iter(0, 0))
        [[]]
    """
    # "bad" input
    if n < 0 or k < 0:
        return

    # Check some corner cases first
    if not k:
        if not n:
            yield []
        return
    n = Integer(n)
    if k == 1:
        yield [n]
        return

    zero = ZZ.zero()
    one = ZZ.one()
    k = int(k)

    pos = 0  # Current position
    rem = zero  # Amount remaining
    cur = [n] + [zero] * (k - 1)  # Current list
    yield list(cur)
    while pos >= 0:
        if not cur[pos]:
            pos -= 1
            continue
        cur[pos] -= one
        rem += one
        if not rem:
            yield list(cur)
        elif pos == k - 2:
            cur[pos + 1] = rem
            yield list(cur)
            cur[pos + 1] = zero
        else:
            pos += 1
            cur[pos] = rem  # Guaranteed to be at least 1
            rem = zero
            yield list(cur)


# October 2012: fixing outdated pickles which use classes being deprecated
from sage.misc.persist import register_unpickle_override
register_unpickle_override('sage.combinat.integer_vector', 'IntegerVectors_nconstraints', IntegerVectorsConstraints)
register_unpickle_override('sage.combinat.integer_vector', 'IntegerVectors_nkconstraints', IntegerVectorsConstraints)



# sage_setup: distribution = sagemath-categories
r"""
Points of bounded height in projective spaces

This module defines functions to compute points of bounded height of a given
number field with height less than a specified bound in projective spaces.

Sage functions to list all elements of a given number field with height less
than a specified bound.

AUTHORS:

- Jing Guo (2022): initial version based on David Krumm's code

REFERENCES:

- [Krumm2016]
"""

import itertools

from math import floor

from sage.arith.functions import lcm
from sage.arith.misc import gcd
from sage.misc.lazy_import import lazy_import
from sage.rings.integer import Integer
from sage.rings.rational_field import QQ
from sage.schemes.projective.projective_space import ProjectiveSpace

lazy_import('sage.geometry.polyhedron.constructor', 'Polyhedron')
lazy_import('sage.libs.pari.all', 'pari')
lazy_import('sage.matrix.constructor', ['matrix', 'column_matrix'])
lazy_import('sage.modules.free_module_element', 'vector')
lazy_import('sage.rings.number_field.unit_group', 'UnitGroup')
lazy_import('sage.rings.real_mpfr', 'RealField')


def ZZ_points_of_bounded_height(PS, dim, bound):
    r"""
    Return an iterator of the points in ``self`` of absolute multiplicative
    height of at most ``bound`` in the rational field.

    INPUT:

    - ``PS`` -- a projective space

    - ``dim`` -- positive integer

    - ``bound`` -- positive integer

    OUTPUT: an iterator of points of bounded height

    EXAMPLES::

        sage: from sage.schemes.projective.proj_bdd_height import ZZ_points_of_bounded_height
        sage: PS = ProjectiveSpace(ZZ, 1)
        sage: sorted(list(ZZ_points_of_bounded_height(PS, 1, 1)))
        [(-1 : -1), (-1 : 0), (-1 : 1), (0 : -1)]
        sage: len(list(ZZ_points_of_bounded_height(PS, 1, 5)))
        40
        sage: sorted(list(ZZ_points_of_bounded_height(PS, 1, 2)))
        [(-2 : -1), (-2 : 1), (-1 : -2), (-1 : -1),
         (-1 : 0), (-1 : 1), (-1 : 2), (0 : -1)]
        sage: PS = ProjectiveSpace(ZZ, 2)
        sage: sorted(list(ZZ_points_of_bounded_height(PS, 2, 1)))
        [(-1 : -1 : -1), (-1 : -1 : 0), (-1 : -1 : 1), (-1 : 0 : -1),
         (-1 : 0 : 0), (-1 : 0 : 1), (-1 : 1 : -1), (-1 : 1 : 0),
         (-1 : 1 : 1), (0 : -1 : -1), (0 : -1 : 0), (0 : -1 : 1),
         (0 : 0 : -1)]

    There are no points of negative height::

        sage: from sage.schemes.projective.proj_bdd_height import ZZ_points_of_bounded_height
        sage: PS = ProjectiveSpace(ZZ, 1)
        sage: list(ZZ_points_of_bounded_height(PS, 1, -3))
        []
    """
    if bound < 1:
        return iter(set())

    points_of_bounded_height = set()

    for t in itertools.product(range(-bound, bound+1), repeat=dim+1):
        if gcd(t) == 1:
            point = PS(t)
            if point not in points_of_bounded_height:
                points_of_bounded_height.add(point)
                yield point


def QQ_points_of_bounded_height(PS, dim, bound, normalize=False):
    r"""
    Return an iterator of the points in ``self`` of absolute multiplicative
    height of at most ``bound`` in the rational field.

    INPUT:

    - ``PS`` -- a projective space

    - ``dim`` -- positive integer

    - ``bound`` -- a real number

    - ``normalize`` -- boolean (default: ``False``); whether to
      normalize the coordinates of returned points

    OUTPUT: an iterator of points of bounded height

    EXAMPLES::

        sage: from sage.schemes.projective.proj_bdd_height import QQ_points_of_bounded_height
        sage: PS = ProjectiveSpace(QQ, 1)
        sage: sorted(list(QQ_points_of_bounded_height(PS, 1, 1)))
        [(-1 : 1), (0 : 1), (1 : 0), (1 : 1)]
        sage: len(list(QQ_points_of_bounded_height(PS, 1, 5)))
        40
        sage: sorted(list(QQ_points_of_bounded_height(PS, 1, 2)))
        [(-2 : 1), (-1 : 1), (-1/2 : 1), (0 : 1),
         (1/2 : 1), (1 : 0), (1 : 1), (2 : 1)]
        sage: sorted(list(QQ_points_of_bounded_height(PS, 1, 2, normalize=True)))
        [(-2 : 1), (-1 : 1), (-1 : 2), (0 : 1),
         (1 : 0), (1 : 1), (1 : 2), (2 : 1)]

    There are no points of negative height::

        sage: from sage.schemes.projective.proj_bdd_height import QQ_points_of_bounded_height
        sage: PS = ProjectiveSpace(QQ, 1)
        sage: list(QQ_points_of_bounded_height(PS, 1, -3))
        []
    """
    if bound < 1:
        return iter(set())

    unit_tuples = list(itertools.product([-1, 1], repeat=dim))
    points_of_bounded_height = set()
    increasing_tuples = itertools.combinations_with_replacement(range(floor(bound + 1)), dim + 1)
    for t in increasing_tuples:
        if gcd(t) == 1:
            for p in itertools.permutations(t):
                for u in unit_tuples:
                    point = PS([a*b for a, b in zip(u, p)] + [p[dim]])
                    if point not in points_of_bounded_height:
                        if normalize:
                            point.scale_by(lcm([point[i].denominator() for i in range(dim + 1)]))

                        points_of_bounded_height.add(point)
                        yield point


def IQ_points_of_bounded_height(PS, K, dim, bound):
    r"""
    Return an iterator of the points in ``self`` of absolute multiplicative
    height of at most ``bound`` in the imaginary quadratic field ``K``.

    INPUT:

    - ``PS`` -- a projective space

    - ``K`` -- a number field

    - ``dim`` -- a positive interger

    - ``bound`` -- a real number

    OUTPUT: an iterator of points of bounded height

    EXAMPLES:

        sage: # needs sage.rings.number_field
        sage: from sage.schemes.projective.proj_bdd_height import IQ_points_of_bounded_height
        sage: CF.<a> = CyclotomicField(3)
        sage: P.<x,y,z> = ProjectiveSpace(CF, 2)
        sage: len(list(IQ_points_of_bounded_height(P, CF, 2, -1)))
        0
        sage: len(list(IQ_points_of_bounded_height(P, CF, 2, 1)))
        57
    """
    if bound < 1:
        return iter([])

    unit_tuples = list(itertools.product(K.roots_of_unity(), repeat=dim))

    class_group_ideals = [c.ideal() for c in K.class_group()]
    class_group_ideal_norms = [i.norm() for i in class_group_ideals]
    class_number = len(class_group_ideals)

    possible_norm_set = set()
    for i in range(class_number):
        for k in range(1, floor(bound + 1)):
            possible_norm_set.add(k*class_group_ideal_norms[i])

    coordinate_space = {}
    coordinate_space[0] = [K(0)]
    for m in possible_norm_set:
        coordinate_space[m] = K.elements_of_norm(m)

    for i in range(class_number):
        a = class_group_ideals[i]
        a_norm = class_group_ideal_norms[i]
        a_norm_bound = bound * a_norm
        a_coordinates = []

        for m in coordinate_space:
            if m <= a_norm_bound:
                for x in coordinate_space[m]:
                    if x in a:
                        a_coordinates.append(x)

        points_in_class_a = set()
        t = len(a_coordinates) - 1
        increasing_tuples = itertools.combinations_with_replacement(range(t + 1), dim + 1)
        for index_tuple in increasing_tuples:
            point_coordinates = [a_coordinates[i] for i in index_tuple]
            if a == K.ideal(point_coordinates):
                for p in itertools.permutations(point_coordinates):
                    for u in unit_tuples:
                        point = PS([i*j for i, j in zip(u, p)] + [p[dim]])

                        if point not in points_in_class_a:
                            points_in_class_a.add(point)
                            yield point


def points_of_bounded_height(PS, K, dim, bound, prec=53):
    r"""
    Return an iterator of the points in ``K`` with dimension ``dim`` of
    absolute multiplicative height of at most ``bound``.

    ALGORITHM:

    This is an implementation of Algorithm 6 in [Krumm2016]_.

    INPUT:

    - ``PS`` -- a projective space

    - ``K`` -- a number field

    - ``dim`` -- positive integer

    - ``bound`` -- a real number

    - ``prec`` -- (default: 53) a positive integer

    OUTPUT: an iterator of points of bounded height

    EXAMPLES::

        sage: from sage.schemes.projective.proj_bdd_height import points_of_bounded_height
        sage: x = polygen(ZZ, 'x')

        sage: # needs sage.geometry.polyhedron sage.libs.pari sage.rings.number_field
        sage: K.<a> = NumberField(x^3 - 7)
        sage: P.<x,y,z> = ProjectiveSpace(K, 2)
        sage: sorted(list(points_of_bounded_height(P, K, 2, 1)))
        [(0 : 0 : 1), (0 : 1 : 0), (1 : 0 : 0), (0 : -1 : 1), (0 : 1 : 1),
         (-1 : 0 : 1), (1 : 0 : 1), (1 : 1 : 0), (-1 : 1 : 0), (-1 : -1 : 1),
         (-1 : 1 : 1), (1 : -1 : 1), (1 : 1 : 1)]

    ::

        sage: # needs sage.geometry.polyhedron sage.libs.pari sage.rings.number_field
        sage: R.<x> = QQ[]
        sage: K.<a> = NumberField(3*x^2 + 1)
        sage: O = K.maximal_order()
        sage: P.<z,w> = ProjectiveSpace(O, 1)
        sage: len(list(P.points_of_bounded_height(bound=2)))
        44

    ::

        sage: # needs sage.geometry.polyhedron sage.libs.pari sage.rings.number_field
        sage: R.<x> = QQ[]
        sage: K.<a> = NumberField(3*x^2 + 1)
        sage: O = K.maximal_order()
        sage: P.<z,w> = ProjectiveSpace(O, 1)
        sage: sorted(list(P.points_of_bounded_height(bound=1)))
        [(-1 : 1), (-3/2*a - 1/2 : 1), (3/2*a - 1/2 : 1), (0 : 1),
         (-3/2*a + 1/2 : 0), (-3/2*a + 1/2 : 1), (3/2*a + 1/2 : 1), (1 : 1)]

    ::

        sage: # needs sage.geometry.polyhedron sage.libs.pari sage.rings.number_field
        sage: R.<x> = QQ[]
        sage: K.<z> = NumberField(x^2 - 2)
        sage: R2.<y> = K[]
        sage: L.<w> = K.extension(y^2 - 3)
        sage: P.<a,b> = ProjectiveSpace(L, 1)
        sage: len(list(P.points_of_bounded_height(bound=2)))
        256
    """
    if bound < 1:
        return iter([])

    r1, r2 = K.signature()
    r = r1 + r2 - 1

    if K.is_relative():
        K_degree = K.relative_degree()
    else:
        K_degree = K.degree()

    roots_of_unity = K.roots_of_unity()
    unit_tuples = list(itertools.product(roots_of_unity, repeat=dim))

    log_embed = K.logarithmic_embedding()

    Reals = RealField(prec)
    logB = Reals(bound).log()

    class_group_ideals = [c.ideal() for c in K.class_group()]
    class_number = len(class_group_ideals)

    if K.is_relative():
        class_group_ideal_norms = [i.absolute_norm() for i in class_group_ideals]
    else:
        class_group_ideal_norms = [i.norm() for i in class_group_ideals]

    norm_bound = bound * max(class_group_ideal_norms)
    fundamental_units = UnitGroup(K).fundamental_units()
    fund_unit_logs = list(map(log_embed, fundamental_units))
    mat = column_matrix(fund_unit_logs)

    test_matrix = mat
    try:
        test_matrix.change_ring(QQ)
    except ValueError:
        raise ValueError('prec too low.')

    cut_fund_unit_logs = mat.delete_rows([r])
    lll_fund_units = []
    for c in pari(cut_fund_unit_logs).qflll().python():
        new_unit = 1
        for i in range(r):
            new_unit *= fundamental_units[i]**c[i]
        lll_fund_units.append(new_unit)
    fundamental_units = lll_fund_units
    fund_unit_logs = list(map(log_embed, fundamental_units))

    possible_norm_set = set()
    for i in range(class_number):
        for k in range(1, floor(bound + 1)):
            possible_norm_set.add(k*class_group_ideal_norms[i])

    principal_ideal_gens = {}
    negative_norm_units = K.elements_of_norm(-1)
    if len(negative_norm_units) == 0:
        for m in possible_norm_set:
            principal_ideal_gens[m] = K.elements_of_norm(m) + K.elements_of_norm(-m)
    else:
        for m in possible_norm_set:
            principal_ideal_gens[m] = K.elements_of_norm(m)

    pr_ideal_gen_logs = {}
    for key in principal_ideal_gens:
        for y in principal_ideal_gens[key]:
            pr_ideal_gen_logs[y] = log_embed(y)

    fund_parallelotope_vertices = []
    for coefficient_tuple in itertools.product([-1/2, 1/2], repeat=r):
        vertex = sum([coefficient_tuple[i]*fund_unit_logs[i] for i in range(r)])
        fund_parallelotope_vertices.append(vertex)

    D_numbers = []
    for v in range(r + 1):
        D_numbers.append(max([vertex[v] for vertex in fund_parallelotope_vertices]))

    A_numbers = []
    for v in range(r + 1):
        A_numbers.append(min([pr_ideal_gen_logs[y][v] for y in pr_ideal_gen_logs]))

    aux_constant = (1/K_degree) * Reals(norm_bound).log()

    L_numbers = []
    for v in range(r1):
        L_numbers.append(aux_constant + D_numbers[v] - A_numbers[v])
    for v in range(r1, r + 1):
        L_numbers.append(2*aux_constant + D_numbers[v] - A_numbers[v])
    L_numbers = vector(L_numbers).change_ring(QQ)

    T = column_matrix(fund_unit_logs).delete_rows([r]).change_ring(QQ)

    # insert_row only takes integers, see https://github.com/sagemath/sage/issues/11328
    M = ((-1)*matrix.identity(r)).insert_row(r, [Integer(1) for i in range(r)])
    M = M.transpose().insert_row(0, [Integer(0) for i in range(r + 1)]).transpose()
    M = M.change_ring(QQ)
    M.set_column(0, L_numbers)
    vertices = map(vector, Polyhedron(ieqs=list(M)).vertices())

    T_it = T.inverse().transpose()
    unit_polytope = Polyhedron([v*T_it for v in vertices])

    coordinate_space = {}
    coordinate_space[0] = [[K(0), log_embed(0)]]
    int_points = unit_polytope.integral_points()

    units_with_logs = {}
    for n in int_points:
        new_unit = 1
        for j in range(r):
            new_unit *= fundamental_units[j]**n[j]
        new_unit_log = sum([n[j]*fund_unit_logs[j] for j in range(r)])
        units_with_logs[n] = [new_unit, new_unit_log]

    for norm in principal_ideal_gens:
        coordinate_list = []
        for y in principal_ideal_gens[norm]:
            for n in int_points:
                unit, unit_log = units_with_logs[n]
                y_log = pr_ideal_gen_logs[y]
                g_log = unit_log + y_log
                bool1 = all(g_log[i] <= aux_constant + D_numbers[i] for i in range(r1))
                bool2 = all(g_log[j] <= 2 * aux_constant + D_numbers[j] for j in range(r1, r + 1))
                if bool1 and bool2:
                    g = unit * y
                    coordinate_list.append([g, g_log])
        if len(coordinate_list) > 0:
            coordinate_space[norm] = coordinate_list

    for m in range(class_number):
        a = class_group_ideals[m]
        a_norm = class_group_ideal_norms[m]
        log_a_norm = Reals(a_norm).log()
        a_const = (logB + log_a_norm)/K_degree
        a_coordinates = []

        for k in range(floor(bound + 1)):
            norm = k * a_norm
            if norm in coordinate_space:
                for pair in coordinate_space[norm]:
                    g, g_log = pair
                    if g in a:
                        bool1 = all(g_log[i] <= a_const + D_numbers[i] for i in range(r1))
                        bool2 = all(g_log[j] <= 2 * a_const + D_numbers[j] for j in range(r1, r + 1))
                        if bool1 and bool2:
                            a_coordinates.append(pair)

        t = len(a_coordinates) - 1
        points_in_class_a = set()
        increasing_tuples = itertools.combinations_with_replacement(range(t + 1), dim + 1)
        log_arch_height_bound = logB + log_a_norm
        for index_tuple in increasing_tuples:
            point_coordinates = [a_coordinates[i][0] for i in index_tuple]
            point_coordinate_logs = [a_coordinates[i][1] for i in index_tuple]
            log_arch_height = sum([max([x[i] for x in point_coordinate_logs]) for i in range(r + 1)])
            if log_arch_height <= log_arch_height_bound and a == K.ideal(point_coordinates):
                for p in itertools.permutations(point_coordinates):
                    for u in unit_tuples:
                        point = PS([i*j for i, j in zip(u, p)] + [p[dim]])

                        if point not in points_in_class_a:
                            points_in_class_a.add(point)
                            yield point



# sage_setup: distribution = sagemath-categories
"""
Dancing links C++ wrapper
"""
# ****************************************************************************
#       Copyright (C) 2008 Carlo Hamalainen <carlo.hamalainen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

# OneExactCover and AllExactCovers are almost exact copies of the
# functions with the same name in sage/combinat/dlx.py by Tom Boothby.

from .dancing_links import dlx_solver


def DLXCPP(rows):
    """
    Solve the Exact Cover problem by using the Dancing Links algorithm
    described by Knuth.

    Consider a matrix M with entries of 0 and 1, and compute a subset
    of the rows of this matrix which sum to the vector of all 1s.

    The dancing links algorithm works particularly well for sparse
    matrices, so the input is a list of lists of the form::

       [
        [i_11,i_12,...,i_1r]
        ...
        [i_m1,i_m2,...,i_ms]
       ]

    where M[j][i_jk] = 1.

    The first example below corresponds to the matrix::

       1110
       1010
       0100
       0001

    which is exactly covered by::

       1110
       0001

    and

    ::

       1010
       0100
       0001

    If soln is a solution given by DLXCPP(rows) then

    [ rows[soln[0]], rows[soln[1]], ... rows[soln[len(soln)-1]] ]

    is an exact cover.

    Solutions are given as a list.

    EXAMPLES::

        sage: rows = [[0,1,2]]
        sage: rows+= [[0,2]]
        sage: rows+= [[1]]
        sage: rows+= [[3]]
        sage: [x for x in DLXCPP(rows)]
        [[3, 0], [3, 1, 2]]
    """
    if not rows:
        return

    x = dlx_solver(rows)

    while x.search():
        yield x.get_solution()


def AllExactCovers(M):
    """
    Solve the exact cover problem on the matrix M (treated as a dense
    binary matrix).

    EXAMPLES: No exact covers::

        sage: M = Matrix([[1,1,0],[1,0,1],[0,1,1]])                                     # needs sage.modules
        sage: [cover for cover in AllExactCovers(M)]                                    # needs sage.modules
        []

    Two exact covers::

        sage: M = Matrix([[1,1,0],[1,0,1],[0,0,1],[0,1,0]])                             # needs sage.modules
        sage: [cover for cover in AllExactCovers(M)]                                    # needs sage.modules
        [[(1, 1, 0), (0, 0, 1)], [(1, 0, 1), (0, 1, 0)]]
    """
    rows = []
    for R in M.rows():
        row = []
        for i in range(len(R)):
            if R[i]:
                row.append(i)
        rows.append(row)
    for s in DLXCPP(rows):
        yield [M.row(i) for i in s]


def OneExactCover(M):
    """
    Solve the exact cover problem on the matrix M (treated as a dense
    binary matrix).

    EXAMPLES::

        sage: # needs sage.modules
        sage: M = Matrix([[1,1,0],[1,0,1],[0,1,1]])  # no exact covers
        sage: print(OneExactCover(M))
        None
        sage: M = Matrix([[1,1,0],[1,0,1],[0,0,1],[0,1,0]]) # two exact covers
        sage: OneExactCover(M)
        [(1, 1, 0), (0, 0, 1)]
    """

    for s in AllExactCovers(M):
        return s



# sage_setup: distribution = sagemath-categories
"""
Reference Parallel Primitives

These are reference implementations of basic parallel
primitives. These are not actually parallel, but work the same way.
They are good for testing.
"""

from sage.misc.prandom import shuffle


def parallel_iter(f, inputs):
    """
    Reference parallel iterator implementation.

    INPUT:

    - ``f`` -- a Python function that can be pickled using
      the ``pickle_function`` command

    - ``inputs`` -- list of pickleable pairs (args, kwds), where
      args is a tuple and kwds is a dictionary

    OUTPUT: iterator over 2-tuples ``(inputs[i], f(inputs[i]))``, where the
    order may be completely random

    EXAMPLES::

        sage: def f(N, M=10): return N*M
        sage: inputs = [((2,3),{}),  (tuple(), {'M':5,'N':3}), ((2,),{})]
        sage: set_random_seed(0)
        sage: for a, val in sage.parallel.reference.parallel_iter(f, inputs):
        ....:     print((a, val))
        (((2,), {}), 20)
        (((), {'M': 5, 'N': 3}), 15)
        (((2, 3), {}), 6)
        sage: for a, val in sage.parallel.reference.parallel_iter(f, inputs):
        ....:     print((a, val))
        (((), {'M': 5, 'N': 3}), 15)
        (((2,), {}), 20)
        (((2, 3), {}), 6)
    """
    v = list(inputs)
    shuffle(v)
    for args, kwds in v:
        yield ((args, kwds), f(*args, **kwds))



# sage_setup: distribution = sagemath-categories
r"""
Points for products of projective spaces

This class builds on the projective space class and its point and morphism classes.

EXAMPLES:

We construct products projective spaces of various dimensions over the same ring.::

    sage: P1xP1.<x,y, u,v> = ProductProjectiveSpaces(QQ, [1, 1])
    sage: P1xP1([2, 1, 3, 1])
    (2 : 1 , 3 : 1)
"""
# ****************************************************************************
# Copyright (C) 2014 Volker Braun <vbraun.name@gmail.com>
#                    Ben Hutz <bn4941@gmail.com>
#
# Distributed under the terms of the GNU General Public License (GPL)
# as published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
# https://www.gnu.org/licenses/
# ****************************************************************************
from copy import copy

import sage.rings.abc

from sage.categories.integral_domains import IntegralDomains
from sage.categories.number_fields import NumberFields
from sage.rings.fraction_field import FractionField
from sage.rings.integer_ring import ZZ
from sage.schemes.generic.morphism import SchemeMorphism
from sage.schemes.generic.morphism import SchemeMorphism_point
from sage.structure.sequence import Sequence
from sage.structure.richcmp import richcmp


class ProductProjectiveSpaces_point_ring(SchemeMorphism_point):
    r"""
    The class of points on products of projective spaces.

    The components are projective space points.

    EXAMPLES::

        sage: T.<x,y,z,w,u> = ProductProjectiveSpaces([2, 1], QQ)
        sage: T.point([1, 2, 3, 4, 5])
        (1/3 : 2/3 : 1 , 4/5 : 1)
    """
    def __init__(self, parent, polys, check=True):
        r"""
        The Python constructor.

        INPUT:

        - ``parent`` -- Hom-set

        - ``polys`` -- anything that defines a point in the class

        - ``check`` -- boolean (default: ``True``); whether or not to perform
          input checks

        EXAMPLES::

            sage: P1.<x0,x1,x2> = ProjectiveSpace(QQ, 2)
            sage: P2 = ProjectiveSpace(QQ, 3, 'y')
            sage: T = ProductProjectiveSpaces([P1, P2])
            sage: Q1 = P1(1, 1, 1)
            sage: Q2 = P2(1, 2, 3, 4)
            sage: T([Q1, Q2])
            (1 : 1 : 1 , 1/4 : 1/2 : 3/4 : 1)

        ::

            sage: T = ProductProjectiveSpaces([2, 2, 2], GF(5), 'x')
            sage: T.point([1, 2, 3, 4, 5, 6, 7, 8, 9])
            (2 : 4 : 1 , 4 : 0 : 1 , 3 : 2 : 1)

        ::

            sage: T.<x,y,z,w> = ProductProjectiveSpaces([1, 1], GF(5))
            sage: X = T.subscheme([x - y, z - 2*w])
            sage: X([1, 1, 2, 1])
            (1 : 1 , 2 : 1)
        """
        polys = copy(polys)
        SchemeMorphism.__init__(self, parent)
        if all(isinstance(P, SchemeMorphism_point) for P in polys):
            if check:
                Q = []
                self._points = []
                for i in range(len(polys)):
                    if polys[i].codomain() != parent.codomain().ambient_space()[i]:
                        raise ValueError("points must be in correct projective spaces")
                    Q += list(polys[i])
                    self._points.append(polys[i])
                parent.codomain()._check_satisfies_equations(Q)
            self._points = polys
        else:
            R = parent.codomain().ambient_space().base_ring()
            polys = Sequence(polys, R)
            N = parent.codomain().ambient_space().dimension_relative_components()
            if check:
                parent.codomain()._check_satisfies_equations(polys)
            splitpolys = self.codomain().ambient_space()._factors(polys)
            self._points = [parent.codomain().ambient_space()[i].point(splitpolys[i], check) for i in range(len(N))]

    def __getitem__(self, i):
        r"""
        Return the ``i``-th coordinate point.

        INPUT:

        - ``i`` -- integer

        OUTPUT: the projective space point that is the ``i``-th coordinate

        EXAMPLES::

            sage: T = ProductProjectiveSpaces([2, 2, 2], GF(5), 'x')
            sage: P = T([1, 0, 1, 1, 0, 0, 0, 0, 1])
            sage: P[1]
            (1 : 0 : 0)
            sage: P[1].codomain()
            Projective Space of dimension 2 over Finite Field of size 5
            sage: P[1][0]
            1
        """
        return self._points[i]

    def _repr_(self):
        r"""
        Return a string representation of this point.

        OUTPUT: string

        EXAMPLES::

            sage: T = ProductProjectiveSpaces([2, 2], ZZ, 'x')
            sage: P = T([1, 2, 3, 4, 5, 6])
            sage: P._repr_()
            '(1 : 2 : 3 , 4 : 5 : 6)'
        """
        return '(%s)' % (" , ".join((" : ".join(repr(f) for f in Q))
                                    for Q in self._points))

    def _richcmp_(self, right, op):
        r"""
        Compare two points in products of projective spaces.

        INPUT:

        - ``other`` -- another point

        OUTPUT: boolean

        EXAMPLES::

            sage: T = ProductProjectiveSpaces([2, 2], ZZ, 'x')
            sage: P = T([1, 2, 3, 4, 5, 6])
            sage: Q = T([2, 4, 6, 4, 5, 6])
            sage: P == Q
            True

        EXAMPLES::

            sage: T = ProductProjectiveSpaces([1, 1, 1], ZZ, 'x')
            sage: P = T([1, 2, 3, 4, 5, 6])
            sage: Q = T([2, 4, 6, 4, 1, 0])
            sage: P != Q
            True

        EXAMPLES::

            sage: T = ProductProjectiveSpaces([1, 1, 1], GF(5), 'x')
            sage: P = T([3, 2, 3, 4, 1, 0])
            sage: Q = T([1, 2, 3, 4, 3, 1])
            sage: P > Q
            True

        ::

            sage: T = ProductProjectiveSpaces([1, 1, 1], GF(5), 'x')
            sage: P = T([1, 2, 3, 4, 1, 0])
            sage: Q = T([1, 2, 3, 4, 3, 0])
            sage: P == Q
            True

        ::

            sage: T = ProductProjectiveSpaces([1, 1, 1], GF(5), 'x')
            sage: P = T([1, 2, 3, 4, 1, 0])
            sage: Q = T([1, 2, 3, 4, 3, 1])
            sage: P < Q
            True
        """
        #needed for Digraph
        if not isinstance(right, (ProductProjectiveSpaces_point_ring)):
            return NotImplemented
        else:
            return richcmp(self._points, right._points, op)

    def __copy__(self):
        r"""
        Return a copy of this point.

        OUTPUT: a point in the same space as third point

        EXAMPLES::

            sage: T = ProductProjectiveSpaces([1, 1], QQ, 'x')
            sage: P = T([2, 1, 0, 1])
            sage: Q = P.__copy__()
            sage: P is Q
            False
            sage: P == Q
            True
        """
        P = [copy(self[i]) for i in range(self.codomain().ambient_space().num_components())]
        return (self.codomain().point(P, False))

    def __iter__(self):
        r"""
        Iterate over the coordinates of the point.

        OUTPUT: an iterator

        EXAMPLES::

            sage: T = ProductProjectiveSpaces([1, 1], QQ, 'x')
            sage: P = T([2, 1, 0, 1])
            sage: iter = P.__iter__()
            sage: next(iter)
            2
            sage: next(iter)
            1
            sage: list(P)
            [2, 1, 0, 1]
        """
        return (x for P in self._points for x in P._coords)

    def __len__(self):
        """
        Return the total number of coordinates in ``self``.

        EXAMPLES::

            sage: T = ProductProjectiveSpaces([1, 1], QQ, 'x')
            sage: P = T([2, 1, 0, 1])
            sage: len(P)
            4
        """
        image = self.codomain().ambient_space()
        return image.dimension() + image.num_components()

    def __hash__(self):
        """
        Compute the hash value of this point.

        OUTPUT: integer

        EXAMPLES::

            sage: PP = ProductProjectiveSpaces(Zmod(6), [1, 1])
            sage: H = hash(PP([5, 1, 3, 4]))

        ::

            sage: PP = ProductProjectiveSpaces(ZZ, [1, 2])
            sage: hash(PP([1, 1, 2, 2, 2])) == hash(PP([1, 1, 1, 1, 1]))
            True

        ::

            sage: PP = ProductProjectiveSpaces(QQ, [1, 1])
            sage: hash(PP([1/7, 1, 2, 1])) == hash((1/7, 1, 2, 1))
            True

        ::

            sage: PP = ProductProjectiveSpaces(GF(7), [1, 1, 1])
            sage: hash(PP([4, 1, 5, 4, 6, 1])) == hash((4, 1, 5, 4, 6, 1))
            False
            sage: hash(PP([4, 1, 5, 4, 6, 1])) == hash((4, 1, 3, 1, 6, 1))
            True
        """
        R = self.codomain().base_ring()
        # if there is a fraction field normalize the point so that
        # equal points have equal hash values
        if R in IntegralDomains():
            P = self.change_ring(FractionField(R))
            P.normalize_coordinates()
            return hash(tuple(P))
        # if there is no good way to normalize return
        # a constant value
        return hash(self.codomain())

    def normalize_coordinates(self):
        r"""
        Remove common factors (componentwise) from the coordinates of this point (including `-1`).

        OUTPUT: none

        EXAMPLES::

            sage: T.<x,y,z,u,v,w> = ProductProjectiveSpaces([2, 2], ZZ)
            sage: P = T.point([5, 10, 15, 4, 2, 6]);
            sage: P.normalize_coordinates()
            sage: P
            (1 : 2 : 3 , 2 : 1 : 3)
        """
        for i in range(self.codomain().ambient_space().num_components()):
            self[i].normalize_coordinates()

    def dehomogenize(self, L):
        r"""
        Dehomogenize `k`-th point at `L[k]`-th coordinate.

        This function computes the appropriate affine patch using ``L``
        and then returns the dehomogenized point on of this affine space.

        INPUT:

        - ``L`` -- list of nonnegative integers

        OUTPUT: :class:`SchemeMorphism_point_affine`

        EXAMPLES::

            sage: PP = ProductProjectiveSpaces([2, 2, 2], QQ, 'x')
            sage: A = PP([2, 4, 6, 23, 46, 23, 9, 3, 1])
            sage: A.dehomogenize([0, 1, 2])
            (2, 3, 1/2, 1/2, 9, 3)

        ::

            sage: # needs sage.rings.real_mpfr sage.symbolic
            sage: PP.<a,b,x,y,z> = ProductProjectiveSpaces([1, 2], CC)
            sage: X = PP.subscheme([a^2 + b^2])
            sage: P = X([2, 2*i, -3, 6*i, 3 - 6*i])
            sage: P.dehomogenize([1,0])
            (-1.00000000000000*I, -2.00000000000000*I, -1.00000000000000 + 2.00000000000000*I)

        ::

            sage: PP = ProductProjectiveSpaces([1, 1], ZZ)
            sage: A = PP([0,1,2,4])
            sage: A.dehomogenize([0,0])
            Traceback (most recent call last):
            ...
            ValueError: can...t dehomogenize at 0 coordinate
        """
        PP = self.codomain()
        A = PP.affine_patch(L)
        pt = []
        for i in range(PP.ambient_space().num_components()):
            pt.extend(self[i].dehomogenize(L[i]))
        return A(pt)

    def scale_by(self, t):
        r"""
        Scale the coordinates of the point by ``t``, done componentwise.

        A :exc:`TypeError` occurs if the point is not in the base ring
        of the codomain after scaling.

        INPUT:

        - ``t`` -- a ring element

        EXAMPLES::

            sage: T.<x, y, z, u, v, w> = ProductProjectiveSpaces([1, 1, 1], ZZ)
            sage: P = T.point([5, 10, 15, 4, 2, 6]);
            sage: P.scale_by([2, 1, 1])
            sage: P
            (10 : 20 , 15 : 4 , 2 : 6)
        """
        if not isinstance(t, (tuple, list)):
            raise TypeError("%s must be a list or tuple" % t)
        if len(t) != self.codomain().ambient_space().num_components():
            raise TypeError("%s must have same number of components as %r" % (t, self))
        for i in range(self.codomain().ambient_space().num_components()):
            self[i].scale_by(t[i])

    def change_ring(self, R, **kwds):
        r"""
        Return a new :class:`ProductProjectiveSpaces_point` which is this point coerced to ``R``.

        If the keyword ``check`` is ``True``, then the initialization checks are performed.
        The user may specify the embedding into ``R`` with a keyword.

        INPUT:

        - ``R`` -- ring

        kwds:

        - ``check`` -- boolean

        - ``embedding`` -- field embedding from the base ring of this point to ``R``

        OUTPUT: :class:`ProductProjectiveSpaces_point`

        EXAMPLES::

            sage: T.<x,y,z,u,v,w> = ProductProjectiveSpaces([1, 1, 1], ZZ)
            sage: P = T.point([5, 3, 15, 4, 2, 6])
            sage: P.change_ring(GF(3))
            (1 : 0 , 0 : 1 , 1 : 0)
        """
        check = kwds.get('check', True)
        S = self.codomain().change_ring(R)
        Q = [P.change_ring(R, **kwds) for P in self._points]
        return S.point(Q, check)

    def global_height(self, prec=None):
        r"""
        Return the absolute logarithmic height of the point.

        This function computes the maximum of global height of each
        component point in the product. Global height of component
        point is computed using function for projective point.

        INPUT:

        - ``prec`` -- desired floating point precision (default:
          default RealField precision)

        OUTPUT: a real number

        EXAMPLES::

            sage: PP = ProductProjectiveSpaces(QQ, [2, 2], 'x')
            sage: Q = PP([1, 7, 5, 18, 2, 3])
            sage: Q.global_height()                                                     # needs sage.symbolic
            2.89037175789616

        ::

            sage: PP = ProductProjectiveSpaces(ZZ, [1, 1], 'x')
            sage: A = PP([-30, 2, 1, 6])
            sage: A.global_height()                                                     # needs sage.symbolic
            2.70805020110221

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = PolynomialRing(QQ)
            sage: k.<w> = NumberField(x^2 + 5)
            sage: PP = ProductProjectiveSpaces(k, [1, 2], 'y')
            sage: Q = PP([3, 5*w + 1, 1, 7*w, 10])
            sage: Q.global_height()
            2.75062910527236

        ::

            sage: PP = ProductProjectiveSpaces(QQbar, [1, 1], 'x')                      # needs sage.rings.number_field
            sage: Q = PP([1, QQbar(sqrt(2)), QQbar(5^(1/3)), QQbar(3^(1/3))])           # needs sage.rings.number_field sage.symbolic
            sage: Q.global_height()                                                     # needs sage.rings.number_field sage.symbolic
            0.536479304144700
        """
        K = self.codomain().base_ring()
        if K not in NumberFields() and K != ZZ and not isinstance(K, (sage.rings.abc.Order, sage.rings.abc.AlgebraicField)):
            raise TypeError("must be over a number field or a number field order or QQbar")

        n = self.codomain().ambient_space().num_components()
        return max(self[i].global_height(prec=prec) for i in range(n))

    def local_height(self, v, prec=None):
        r"""
        Return the maximum of the local height of the coordinates of this point.

        This function computes the maximum of local height of each
        component point in the product. Local height of component
        point is computed using function for projective point.

        INPUT:

        - ``v`` -- a prime or prime ideal of the base ring

        - ``prec`` -- desired floating point precision (default:
          default RealField precision)

        OUTPUT: a real number

        EXAMPLES::

            sage: PP = ProductProjectiveSpaces(QQ, [1, 1], 'x')
            sage: A = PP([11, 5, 10, 2])
            sage: A.local_height(5)                                                     # needs sage.rings.real_mpfr
            1.60943791243410

        ::

            sage: P = ProductProjectiveSpaces(QQ, [1, 2], 'x')
            sage: Q = P([1, 4, 1/2, 2, 32])
            sage: Q.local_height(2)                                                     # needs sage.rings.real_mpfr
            4.15888308335967
        """
        K = FractionField(self.domain().base_ring())
        if K not in NumberFields():
            raise TypeError("must be over a number field or a number field order")

        n = self.codomain().ambient_space().num_components()
        return max(self[i].local_height(v, prec=prec) for i in range(n))


class ProductProjectiveSpaces_point_field(ProductProjectiveSpaces_point_ring):

    def intersection_multiplicity(self, X):
        r"""
        Return the intersection multiplicity of the codomain of this point and subscheme ``X`` at this point.

        This uses the subscheme implementation of intersection_multiplicity. This point must be a point
        on a subscheme of a product of projective spaces.

        INPUT:

        - ``X`` -- a subscheme in the same ambient space as the codomain of this point

        OUTPUT: integer

        EXAMPLES::

            sage: PP.<x,y,z,u,v> = ProductProjectiveSpaces(QQ, [2, 1])
            sage: X = PP.subscheme([y^2*z^3*u - x^5*v])
            sage: Y = PP.subscheme([u^3 - v^3, x - y])
            sage: Q = X([0,0,1,1,1])
            sage: Q.intersection_multiplicity(Y)                                        # needs sage.libs.singular
            2
        """
        from sage.schemes.product_projective.space import ProductProjectiveSpaces_ring
        if isinstance(self.codomain(), ProductProjectiveSpaces_ring):
            raise TypeError("this point must be a point on a subscheme of a product of projective spaces")
        return self.codomain().intersection_multiplicity(X, self)

    def multiplicity(self):
        r"""
        Return the multiplicity of this point on its codomain.

        This uses the subscheme implementation of multiplicity. This point must be a point
        on a subscheme of a product of projective spaces.

        OUTPUT: integer

        EXAMPLES::

            sage: PP.<x,y,z,w,u,v,t> = ProductProjectiveSpaces(QQ, [3, 2])
            sage: X = PP.subscheme([x^8*t - y^8*t + z^5*w^3*v])
            sage: Q1 = X([1,1,0,0,-1,-1,1])
            sage: Q1.multiplicity()                                                     # needs sage.libs.singular
            1
            sage: Q2 = X([0,0,0,1,0,1,1])
            sage: Q2.multiplicity()                                                     # needs sage.libs.singular
            5
            sage: Q3 = X([0,0,0,1,1,0,0])
            sage: Q3.multiplicity()                                                     # needs sage.libs.singular
            6
        """
        from sage.schemes.product_projective.space import ProductProjectiveSpaces_ring
        if isinstance(self.codomain(), ProductProjectiveSpaces_ring):
            raise TypeError("this point must be a point on a subscheme of a product of projective spaces")
        return self.codomain().multiplicity(self)


class ProductProjectiveSpaces_point_finite_field(ProductProjectiveSpaces_point_field):
    pass



# sage_setup: distribution = sagemath-categories
"""
Monoid of ideals in a commutative ring

WARNING: This is used by some rings that are not commutative! ::

    sage: MS = MatrixSpace(QQ, 3, 3)                                                    # needs sage.modules
    sage: type(MS.ideal(MS.one()).parent())                                             # needs sage.modules
    <class 'sage.rings.ideal_monoid.IdealMonoid_c_with_category'>
"""

import sage.rings.integer_ring
from sage.categories.monoids import Monoids
from sage.rings import ideal
from sage.structure.parent import Parent


def IdealMonoid(R):
    r"""
    Return the monoid of ideals in the ring ``R``.

    EXAMPLES::

        sage: R = QQ['x']
        sage: from sage.rings.ideal_monoid import IdealMonoid
        sage: IdealMonoid(R)
        Monoid of ideals of Univariate Polynomial Ring in x over Rational Field
    """
    return IdealMonoid_c(R)


class IdealMonoid_c(Parent):
    r"""
    The monoid of ideals in a commutative ring.

    TESTS::

        sage: R = QQ['x']
        sage: from sage.rings.ideal_monoid import IdealMonoid
        sage: M = IdealMonoid(R)
        sage: TestSuite(M).run()
          Failure in _test_category:
        ...
        The following tests failed: _test_elements

    (The "_test_category" test fails but I haven't the foggiest idea why.)
    """

    Element = ideal.Ideal_generic  # this doesn't seem to do anything

    def __init__(self, R):
        r"""
        Initialize ``self``.

        TESTS::

            sage: R = QuadraticField(-23, 'a')                                          # needs sage.rings.number_field
            sage: from sage.rings.ideal_monoid import IdealMonoid
            sage: M = IdealMonoid(R); M  # indirect doctest                             # needs sage.rings.number_field
            Monoid of ideals of Number Field in a with defining polynomial x^2 + 23
             with a = 4.795831523312720?*I

            sage: id = QQ.ideal(6)
            sage: id.parent().category()
            Category of commutative monoids

            sage: MS = MatrixSpace(QQ, 3, 3)                                            # needs sage.modules
            sage: MS.ideal(MS.one()).parent().category()                                # needs sage.modules
            Category of monoids
        """
        self.__R = R
        cat = Monoids()
        if R.is_commutative():
            cat = cat.Commutative()
        Parent.__init__(self, base=sage.rings.integer_ring.ZZ,
                        category=cat)
        self._populate_coercion_lists_()

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        TESTS::

            sage: R = QuadraticField(-23, 'a')                                          # needs sage.rings.number_field
            sage: from sage.rings.ideal_monoid import IdealMonoid
            sage: M = IdealMonoid(R); M._repr_()                                        # needs sage.rings.number_field
            'Monoid of ideals of Number Field in a with defining polynomial x^2 + 23
             with a = 4.795831523312720?*I'
        """
        return "Monoid of ideals of %s" % self.__R

    def ring(self):
        r"""
        Return the ring of which this is the ideal monoid.

        EXAMPLES::

            sage: R = QuadraticField(-23, 'a')                                          # needs sage.rings.number_field
            sage: from sage.rings.ideal_monoid import IdealMonoid
            sage: M = IdealMonoid(R); M.ring() is R                                     # needs sage.rings.number_field
            True
        """
        return self.__R

    def _element_constructor_(self, x):
        r"""
        Create an ideal in this monoid from ``x``.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R.<a> = QuadraticField(-23)
            sage: from sage.rings.ideal_monoid import IdealMonoid
            sage: M = IdealMonoid(R)
            sage: M(a)   # indirect doctest
            Fractional ideal (a)
            sage: M([a-4, 13])
            Fractional ideal (13, 1/2*a + 9/2)
        """
        try:
            side = x.side()
        except (AttributeError, TypeError):
            side = None
        try:
            x = x.gens()
        except AttributeError:
            pass
        if side is None:
            y = self.__R.ideal(x)
        else:
            y = self.__R.ideal(x, side=side)
        y._set_parent(self)
        return y

    def _coerce_map_from_(self, x):
        r"""
        Used by coercion framework.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R = QuadraticField(-23, 'a')
            sage: M = R.ideal_monoid()
            sage: M.has_coerce_map_from(R) # indirect doctest
            True
            sage: M.has_coerce_map_from(QQ.ideal_monoid())
            True
            sage: M.has_coerce_map_from(Zmod(6))
            False
            sage: M.has_coerce_map_from(loads(dumps(M)))
            True
        """
        if isinstance(x, IdealMonoid_c):
            return self.ring().has_coerce_map_from(x.ring())
        else:
            return self.ring().has_coerce_map_from(x)

    def __eq__(self, other):
        r"""
        Check whether ``self`` is not equal to ``other``.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R = QuadraticField(-23, 'a')
            sage: M = R.ideal_monoid()
            sage: M == QQ
            False
            sage: M == 17
            False
            sage: M == R.ideal_monoid()
            True
        """
        if not isinstance(other, IdealMonoid_c):
            return False
        else:
            return self.ring() == other.ring()

    def __ne__(self, other):
        r"""
        Check whether ``self`` is not equal to ``other``.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R = QuadraticField(-23, 'a')
            sage: M = R.ideal_monoid()
            sage: M != QQ
            True
            sage: M != 17
            True
            sage: M != R.ideal_monoid()
            False
        """
        return not (self == other)

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R = QuadraticField(-23, 'a')
            sage: M = R.ideal_monoid()
            sage: hash(M) == hash(QQ)
            False
            sage: hash(M) == 17
            False
            sage: hash(M) == hash(R.ideal_monoid())
            True
        """
        # uses a random number, to have a distinct hash
        return hash((1580963238588124931699, self.ring()))
